"""
Toolbox
-------

This module contains helper functionality for ``optics_measurements``.
It provides functions to perform regularly used simple calculations.
"""
import numpy as np


def df_diff(df, a_col, b_col):
    """ Returns a column containing the difference between a_col and b_col """
    return df.loc[:, a_col].to_numpy() - df.loc[:, b_col].to_numpy()


def df_sum(df, a_col, b_col):
    """ Returns a column containing the sum of a_col and b_col """
    return df.loc[:, a_col].to_numpy() + df.loc[:, b_col].to_numpy()


def df_ratio(df, a_col, b_col):
    """ Returns a column containing the ratio between a_col and b_col """
    return df.loc[:, a_col].to_numpy() / df.loc[:, b_col].to_numpy()


def df_prod(df, a_col, b_col):
    """ Returns a column containing the product of a_col and b_col """
    return df.loc[:, a_col].to_numpy() * df.loc[:, b_col].to_numpy()


def df_rel_diff(df, a_col, b_col):
    """ Returns a column containing the difference between a_col and b_col relative to b_col """
    return df_ratio(df, a_col, b_col) - 1


# with Errors ---


def df_err_sum(df, a_err_col, b_err_col):
    """ Returns a column containing the root of the sum-of-squares of a_err_col and b_err_col """
    return np.sqrt(np.square(df.loc[:, a_err_col].to_numpy()) + np.square(df.loc[:, b_err_col].to_numpy()))


def df_rel_err_sum(df, a_col, b_col, a_err_col, b_err_col):
    """ Returns a column containing the root of the relative sum-of-square of a_col/a_err_col and b_col/b_err_col """
    return np.sqrt(np.square(df_ratio(df, a_err_col, a_col)) + np.square(df_ratio(df, b_err_col, b_col)))


def df_sum_with_err(df, a_col, b_col, a_err_col, b_err_col):
    """ Returns two columns containing the sum and the total errors of the given columns."""
    return df_sum(df, a_col, b_col), df_err_sum(df, a_err_col, b_err_col)


def df_diff_with_err(df, a_col, b_col, a_err_col, b_err_col):
    """ Returns two columns containing the difference and the total errors of the given columns."""
    return df_diff(df, a_col, b_col), df_err_sum(df, a_err_col, b_err_col)


def df_rel_diff_with_err(df, a_col, b_col, a_err_col, b_err_col):
    """ Returns two columns containing the relative difference and the total errors of the given columns."""
    return (df_rel_diff(df, a_col, b_col),
            np.abs(df_ratio(df, a_col, b_col)) * df_rel_err_sum(df, a_col, b_col, a_err_col, b_err_col))


def df_ratio_with_err(df, a_col, b_col, a_err_col, b_err_col):
    """ Returns two columns containing the ratio and the total errors of the given columns."""
    ratio = df_ratio(df, a_col, b_col)
    return ratio, np.abs(ratio) * df_rel_err_sum(df, a_col, b_col, a_err_col, b_err_col)


def df_prod_with_err(df, a_col, b_col, a_err_col, b_err_col):
    """ Returns two columns containing the product and the total errors of the given columns."""
    prod = df_prod(df, a_col, b_col)
    return prod, np.abs(prod) * df_rel_err_sum(df, a_col, b_col, a_err_col, b_err_col)


# Angular Calculations ---


def df_ang_diff(df, a_col, b_col):
    """ Returns a column containing the angular difference between angles a and b in [-0.5 , 0.5] """
    return ang_diff(df.loc[:, a_col].to_numpy(), df.loc[:, b_col].to_numpy())


def ang_diff(a, b):
    return ang_interval_check(ang_interval_check(a) - ang_interval_check(b))


def ang_sum(a, b):
    """ Returns a column containing the angular sum between angles a and b in [-0.5 , 0.5] """
    return ang_interval_check(ang_interval_check(a) + ang_interval_check(b))


def ang_interval_check(ang):
    """ Returns ang wrapped into [-0.5, 0.5] """
    return np.where(np.abs(ang) > 0.5, ang - np.sign(ang), ang)



"""
Timber Extraction
-----------------

Tools to extract data from ``Timber``. It is a bit heavy on the LHC side at the moment.

**Please note**: this module requires the ``pytimber`` package to access ``Timber`` functionality,
both of which are only possible from inside the CERN network.

To install ``pytimber`` along ``omc3``, please do so from inside the CERN network by using the [cern] extra
dependency and installing from the ``acc-py`` package index (by specifying ``--index-url
https://acc-py-repo.cern.ch/repository/vr-py-releases/simple`` and
``--trusted-host acc-py-repo.cern.ch`` to your ``pip`` installation command).
"""
import datetime
import re
from contextlib import suppress
from typing import Dict, List, NewType, Sequence, Tuple, Union

import numpy as np
import tfs

# from jpype import JException, java
from omc3.tune_analysis import constants as const
from omc3.utils import logging_tools
from omc3.utils.mock import cern_network_import
from omc3.utils.time_tools import CERNDatetime

TIME_COL = const.get_time_col()
START_TIME = const.get_tstart_head()
END_TIME = const.get_tend_head()

LOG = logging_tools.get_logger(__name__)
pytimber = cern_network_import("pytimber")
jpype = cern_network_import("jpype")

MAX_RETRIES = 10  # number of retries on retryable exception
AcceptableTimeStamp = NewType("AcceptableTimeStamp", Union[CERNDatetime, int, float])


def lhc_fill_to_tfs(
    fill_number: int, keys: Sequence[str] = None, names: Dict[str, str] = None
) -> tfs.TfsDataFrame:
    """
    Extracts data for keys of fill from ``Timber``.

    Args:
        fill_number (int): Number of the fill to extract from.
        keys (Sequence[str]): the different variables names to extract data for.
        names (Dict[str, str): dict mapping keys to column names.

    Returns:
        The extracted data as a ``TfsDataFrame``.
    """
    db = pytimber.LoggingDB(source="nxcals")
    t_start, t_end = get_fill_times(db, fill_number)
    return extract_between_times(t_start, t_end, keys, names)


def extract_between_times(
    t_start: AcceptableTimeStamp,
    t_end: AcceptableTimeStamp,
    keys: Sequence[str] = None,
    names: Dict[str, str] = None,
) -> tfs.TfsDataFrame:
    """
    Extracts data for keys between ``t_start`` and ``t_end`` from ``Timber``.

    Args:
        t_start (AcceptableTimeStamp): starting time in CERNDateTime or timestamp.
        t_end (AcceptableTimeStamp): end time in local CERNDateTime or timestamp.
        keys (Sequence[str]): the different variables names to extract data for.
        names (Dict[str, str): dict mapping keys to column names.

    Returns:
        Extracted data in a ``TfsDataFrame``.
    """
    with suppress(TypeError):
        t_start: CERNDatetime = CERNDatetime.from_timestamp(t_start)

    with suppress(TypeError):
        t_end: CERNDatetime = CERNDatetime.from_timestamp(t_end)

    db = pytimber.LoggingDB(source="nxcals")
    if keys is None:
        keys = get_tune_and_coupling_variables(db)

    # Attempt getting data from NXCALS, which can sometimes need a few retries (yay NXCALS)
    # If Java gives a feign.RetryableException, retry up to MAX_RETRIES times.
    extract_dict = {}
    for tries in range(MAX_RETRIES + 1):
        try:
            # We use timestamps to avoid any confusion with local time
            extract_dict = db.get(keys, t_start.timestamp(), t_end.timestamp())
        except jpype.java.lang.IllegalStateException as java_state_error:
            raise IOError(
                "Could not get data from Timber, user probably has no access to NXCALS"
            ) from java_state_error
        except jpype.JException as java_exception:  # Might be a case for retries
            if "RetryableException" in str(java_exception) and (tries + 1) < MAX_RETRIES:
                LOG.warning(f"Could not get data from Timber! Trial no {tries + 1} / {MAX_RETRIES}")
                continue  # will go to the next iteratoin of the loop, so retry
            raise IOError("Could not get data from timber!") from java_exception
        else:
            break

    if (not len(extract_dict)  # dict is empty
            or all(not len(v) for v in extract_dict.values())  # values are empty
            or all(len(v) == 2 and not len(v[0]) for v in extract_dict.values())  # arrays are empty (size 2 for time/data)
    ):
        raise IOError(f"Variables {keys} found but no data extracted in time {t_start.utc_string} - {t_end.utc_string} (UTC).\n"
                      f"Possible reasons:\n"
                      f"  - Too small time window.\n"
                      f"  - Old pytimber version.\n"
                      f"  - Variable outdated (i.e. no longer logged).")

    out_df = tfs.TfsDataFrame()
    for key in keys:
        if extract_dict[key][1][0].size > 1:
            raise NotImplementedError("Multidimensional variables are not implemented yet")

        data = np.asarray(extract_dict[key]).transpose()
        column = key if names is None else names.get(key, key)
        key_df = tfs.TfsDataFrame(data, columns=[TIME_COL, column]).set_index(TIME_COL)
        out_df = out_df.merge(key_df, how="outer", left_index=True, right_index=True)

    out_df.index = [CERNDatetime.from_timestamp(i) for i in out_df.index]
    out_df.headers[START_TIME] = t_start.cern_utc_string()
    out_df.headers[END_TIME] = t_end.cern_utc_string()
    return out_df


def get_tune_and_coupling_variables(db) -> List[str]:
    """
    Returns the tune and coupling variable names.

    Args:
        db (pytimber.LoggingDB): pytimber database connexion.

    Returns:
        List of variable names as strings.
    """
    bbq_vars = []
    for search_term in ["%EIGEN%FREQ%", "%COUPL%ABS%"]:
        search_results = db.search(search_term)
        for res in search_results:
            if re.match(r"LHC\.B(OFSU|QBBQ\.CONTINUOUS)", res):
                bbq_vars.append(res)
    return bbq_vars


def get_fill_times(
    db, fill_number: int
) -> Tuple[Union[datetime.datetime, float], Union[datetime.datetime, float]]:
    """
    Returns start and end time of fill with fill number.

    Args:
        db (pytimber.LoggingDB): pytimber database.
        fill_number (int): fill number.

    Returns:
       `Tuple` of start and end time.
    """
    fill = db.getLHCFillData(fill_number)
    return fill["startTime"], fill["endTime"]



"""
Clean
-----

This module contains the cleaning functionality of ``harpy``.
"""
import numpy as np
import pandas as pd

from omc3.utils import logging_tools
from omc3.utils.contexts import timeit

LOGGER = logging_tools.getLogger(__name__)
NTS_LIMIT = 8.  # Noise to signal limit


def clean(harpy_input, bpm_data, model):
    """
    Cleans BPM TbT matrix: removes BPMs not present in the model and based on specified cuts.
    Also cleans the noise using singular value decomposition.

    Args:
        harpy_input: The input object containing the analysis settings.
        bpm_data: DataFrame of BPM TbT matrix indexed by BPM names.
        model: model containing BPMs longitudinal locations indexed by BPM names.

    Returns:
        Clean BPM matrix, its decomposition, bad BPMs summary and estimated BPM resolutions
    """
    bpm_data, bpms_not_in_model = _get_only_model_bpms(bpm_data, model)
    if bpm_data.empty:
        raise AssertionError("Check BPMs names! None of the BPMs were found in the model!")
    if not harpy_input.clean:
        return bpm_data, None, bpms_not_in_model, None
    with timeit(lambda spanned: LOGGER.debug(f"Time for filtering: {spanned}")):
        bpm_data, bad_bpms_clean = _cut_cleaning(harpy_input, bpm_data, model)
    with timeit(lambda spanned: LOGGER.debug(f"Time for SVD clean: {spanned}")):
        bpm_data, bpm_res, bad_bpms_svd, usv = _svd_clean(bpm_data, harpy_input)
        all_bad_bpms = bpms_not_in_model + bad_bpms_clean + bad_bpms_svd
    return bpm_data, usv, all_bad_bpms, bpm_res


def _get_only_model_bpms(bpm_data, model):
    if model is None:
        return bpm_data, []
    bpm_data_in_model = bpm_data.loc[model.index.intersection(bpm_data.index)]
    not_in_model = bpm_data.index.difference(model.index)
    return bpm_data_in_model, [f"{bpm} not found in model" for bpm in not_in_model]


def _cut_cleaning(harpy_input, bpm_data, model):
    LOGGER.debug(f"Number of BPMs in the input {bpm_data.index.size}")
    known_bad_bpms = _detect_known_bad_bpms(bpm_data, harpy_input.bad_bpms)
    bpm_flatness = _detect_flat_bpms(bpm_data, harpy_input.peak_to_peak)
    bpm_spikes = _detect_bpms_with_spikes(bpm_data, harpy_input.max_peak)
    exact_zeros = _detect_bpms_with_exact_zeros(bpm_data, harpy_input.keep_exact_zeros)
    all_bad_bpms = _index_union(known_bad_bpms, bpm_flatness, bpm_spikes, exact_zeros)
    original_bpms = bpm_data.index

    bpm_data = bpm_data.loc[bpm_data.index.difference(all_bad_bpms, sort=False)]
    bad_bpms_with_reasons = _get_bad_bpms_summary(
        harpy_input, known_bad_bpms, bpm_flatness, bpm_spikes, exact_zeros
    )
    _report_clean_stats(original_bpms.size, bpm_data.index.size)
    bpm_data = _fix_polarity(harpy_input.wrong_polarity_bpms, bpm_data)
    if model is not None and harpy_input.first_bpm is not None:
        bpm_data = _resync_bpms(harpy_input, bpm_data, model)
    return bpm_data, bad_bpms_with_reasons


def _svd_clean(bpm_data, harpy_input):
    bpm_data_mean = np.mean(bpm_data.to_numpy(), axis=1)
    u_mat, sv_mat, u_mask = svd_decomposition(bpm_data - bpm_data_mean[:, None],
                                              harpy_input.sing_val,
                                              dominance_limit=harpy_input.svd_dominance_limit,
                                              num_iter=harpy_input.num_svd_iterations)

    clean_u, dominant_bpms = _clean_dominant_bpms(u_mat, u_mask, harpy_input.svd_dominance_limit)
    clean_data = clean_u.dot(sv_mat) + bpm_data_mean[np.all(u_mask, axis=1), None]
    bpm_res = (clean_data - bpm_data.loc[clean_u.index]).std(axis=1)
    orbit_offset = (clean_data - bpm_data.loc[clean_u.index]).mean(axis=1)
    LOGGER.debug(f"Average closed orbit offset: {np.mean(orbit_offset)}")
    LOGGER.debug(f"Average BPM resolution: {np.mean(bpm_res)}")
    average_signal = np.mean(np.std(clean_data, axis=1))
    LOGGER.debug(f"np.mean(np.std(A, axis=1): {average_signal}")
    if np.mean(bpm_res) > NTS_LIMIT * average_signal:
        raise ValueError("The data is too noisy. The most probable explanation "
                         "is that there was no excitation or it was very low.")
    return clean_data, bpm_res, dominant_bpms, (clean_u, sv_mat - np.mean(sv_mat, axis=1)[:, None])


def _detect_known_bad_bpms(bpm_data, list_of_bad_bpms):
    """Searches for known bad BPMs."""
    return bpm_data.index.intersection(list_of_bad_bpms)


def _detect_flat_bpms(bpm_data, min_peak_to_peak):
    """Detects BPMs with the same values for all turns."""
    cond = ((bpm_data.max(axis=1) - bpm_data.min(axis=1)).abs() < min_peak_to_peak)
    bpm_flatness = bpm_data[cond].index
    if bpm_flatness.size:
        LOGGER.debug(f"Flat BPMS detected (diff min/max <= {min_peak_to_peak}. "
                     f"BPMs removed: {bpm_flatness.size}")
    return bpm_flatness


def _detect_bpms_with_spikes(bpm_data, max_peak_cut):
    """Detects BPMs with spikes > `max_peak_cut`."""
    too_high = bpm_data[bpm_data.max(axis=1) > max_peak_cut].index
    too_low = bpm_data[bpm_data.min(axis=1) < -max_peak_cut].index
    bpm_spikes = too_high.union(too_low)
    if bpm_spikes.size:
        LOGGER.debug(f"Spikes > {max_peak_cut} detected. BPMs removed: {bpm_spikes.size}")
    return bpm_spikes


def _detect_bpms_with_exact_zeros(bpm_data, keep_exact_zeros):
    """Detects BPMs with exact zeros due to OP workaround."""
    if keep_exact_zeros:
        LOGGER.debug("Skipped exact zero check")
        return pd.Index([])
    exact_zeros = bpm_data[~np.all(bpm_data, axis=1)].index
    if exact_zeros.size:
        LOGGER.debug(f"Exact zeros detected. BPMs removed: {exact_zeros.size}")
    return exact_zeros


def _get_bad_bpms_summary(harpy_input, known_bad_bpms, bpm_flatness, bpm_spikes, exact_zeros):
    return ([f"{bpm_name} Known bad BPM" for bpm_name in known_bad_bpms] +
            [f"{bpm_name} Flat BPM, the difference between min/max is smaller than "
            f"{harpy_input.peak_to_peak}" for bpm_name in bpm_flatness] +
            [f"{bpm_name} Spiky BPM, found spike higher than "
            f"{harpy_input.max_peak}" for bpm_name in bpm_spikes] +
            [f"{bpm_name} Found an exact zero" for bpm_name in exact_zeros])


def _report_clean_stats(n_total_bpms, n_good_bpms):
    LOGGER.debug("Filtering done:")
    if n_total_bpms == 0:
        raise ValueError("Total Number of BPMs after filtering is zero.")
    n_bad_bpms = n_total_bpms - n_good_bpms
    LOGGER.debug(f"(Statistics for file reading) Total BPMs: {n_total_bpms}, "
                 f"Good BPMs: {n_good_bpms} ({(100 * n_good_bpms / n_total_bpms):2.2f}%), "
                 f"Bad BPMs: {n_bad_bpms} ({(100 * n_bad_bpms / n_total_bpms):2.2f}%)")
    if (n_good_bpms / n_total_bpms) < 0.5:
        raise ValueError("More than half of BPMs are bad. "
                         "This could be because a bunch not present in the machine has been "
                         "selected or because of a problem with the phasing of the BPMs.")


def _index_union(*indices):
    new_index = pd.Index([])
    for index in indices:
        new_index = new_index.union(index)
    return new_index


def _fix_polarity(wrong_polarity_names, bpm_data):
    """Fixes wrong polarity."""
    bpm_data.loc[wrong_polarity_names, :] = -1 * bpm_data.loc[wrong_polarity_names, :].to_numpy()
    return bpm_data


def _resync_bpms(harpy_input, bpm_data, model):
    """Resynchronizes BPMs between the injection point and start of the lattice."""
    LOGGER.debug("Will resynchronize BPMs")
    bpm_pos = model.index.get_loc(harpy_input.first_bpm)
    if harpy_input.opposite_direction:
        mask = np.array([x in model.index[bpm_pos::-1] for x in bpm_data.index])
    else:
        mask = np.array([x in model.index[bpm_pos:] for x in bpm_data.index])
    bpm_data.loc[mask] = np.roll(bpm_data.loc[mask], -1, axis=1)
    return bpm_data.iloc[:, :-1]


def svd_decomposition(matrix, num_singular_values, dominance_limit=None, num_iter=None):
    """
    Computes reduced (K largest values) singular value decomposition of a matrix.
    Requiring K singular values from MxN matrix results in matrices sized:
    `((M,K) x diag(K) x (K,N))`

    Args:
        matrix: matrix to be decomposed.
        num_singular_values: Required number of singular values for reconstruction.
        dominance_limit: limit on SVD dominance.
        num_iter: maximal number of iteration to remove elements and renormalise matrices.

    Returns:
        An indexed DataFrame of U matrix `(M,K)`,
        the product of S and V^T matrices `(diag(K).x(K,N))`,
        and the U matrix mask for cleaned elements.
    """
    u_mat, s_mat, vt_mat = np.linalg.svd(matrix, full_matrices=False)
    u_mat, s_mat, u_mat_mask = _remove_dominant_elements(u_mat, s_mat, dominance_limit, num_iter=num_iter)

    available = np.sum(s_mat > 0.)
    if num_singular_values > available:
        LOGGER.warning(f"Requested more singular values than available(={available})")
    keep = min(num_singular_values, available)
    LOGGER.debug(f"Number of singular values to keep: {keep}")

    indices = np.argsort(s_mat)[::-1][:keep]
    return (pd.DataFrame(index=matrix.index, data=u_mat[:, indices]),
            np.dot(np.diag(s_mat[indices]), vt_mat[indices, :]),
            u_mat_mask[:, :int(np.max(indices))+1])


def _remove_dominant_elements(u_mat, s_mat, dominance_limit, num_iter=3):
    u_mat_mask = np.ones(u_mat.shape, dtype=bool)
    if dominance_limit is None:
        return u_mat, s_mat, u_mat_mask
    if dominance_limit < 1 / np.sqrt(2):
        LOGGER.warning(f"The svd_dominance_limit looks too low: {dominance_limit}")

    for i in range(num_iter):
        if np.all(np.abs(u_mat) <= dominance_limit):
            break
        condition = np.logical_and(np.abs(u_mat) > dominance_limit,
                                   np.abs(u_mat) == np.max(np.abs(u_mat), axis=0))
        u_mat_mask[condition] = False
        u_mat[condition] = 0.0
        norms = np.sqrt(np.sum(np.square(u_mat), axis=0))
        u_mat = u_mat / norms
        s_mat = s_mat * norms
    return u_mat, s_mat, u_mat_mask


def _clean_dominant_bpms(u_mat, u_mat_mask, svd_dominance_limit):
    dominant_bpms = u_mat[np.any(~u_mat_mask, axis=1)].index
    if dominant_bpms.size > 0:
        LOGGER.debug(f"Bad BPMs from SVD detected. Number of BPMs removed: {dominant_bpms.size}")
    clean_u = u_mat.loc[u_mat.index.difference(dominant_bpms, sort=False)]
    return clean_u, [f"{bpm_name} Dominant BPM in SVD, peak value > {svd_dominance_limit}"
                     for bpm_name in dominant_bpms]



"""
PS Booster Model Creator
------------------------

This module provides convenience functions for model creation of the ``PSB``.
"""
import shutil

from omc3.model.accelerators.accelerator import AccExcitationMode
from omc3.model.accelerators.psbooster import Psbooster
from omc3.model.constants import ERROR_DEFFS_TXT
from omc3.model.model_creators.abstract_model_creator import ModelCreator


class PsboosterModelCreator(ModelCreator):
    @classmethod
    def get_madx_script(cls, accel: Psbooster) -> str:
        madx_script = accel.get_base_madx_script()
        replace_dict = {
            "USE_ACD": str(int(accel.excitation == AccExcitationMode.ACD)),
            "RING": accel.ring,
            "DPP": accel.dpp,
            "OUTPUT": str(accel.model_dir),
        }
        madx_template = accel.get_file("twiss.mask").read_text()
        madx_script += madx_template % replace_dict
        return madx_script

    @classmethod
    def get_correction_check_script(cls, accel: Psbooster, corr_file: str, chrom: bool) -> str:
        raise NotImplemented(
            "Correction check is not implemented for the PsBooster model creator yet. "
        )

    @classmethod
    def prepare_run(cls, accel: Psbooster):
        shutil.copy(
            accel.get_file(f"error_deff_ring{accel.ring}.txt"), accel.model_dir / ERROR_DEFFS_TXT
        )



"""
Plot Spectrum
-------------

Spectrum plotter for frequency analysis output-data (supports also DRIVE output).

The spectra can be either plotted as `stem`-plots or as `waterfall`-plots.
The stem-plots can be in any combination: split by given files, split by given bpms or combined
in any way (by usage of the `combine_by` option).
Note that if both of those are false (as is default) there will anyway be only one waterfall plot
per given input file.


In case of split-by-file, plots are saved in a sub-directory of the given `output_dir` with the
name of the original TbT file.
In case of split by bpm the plots will have the bpm-name in their filename.


The `lines_tunes` and `lines_nattunes` lists accept tuples of multipliers for the respective
tunes, which define the resonance lines plotted into the spectrum as well. A dashed line will
indicate the average of all tunes given in the data of one figure, while a semi-transparent area
will indicate min- and max- values of this line.

With `lines_manual`, one can plot vertical lines at manual locations (see
parameter specs below).

The function returns two dictionaries, where the first dictionary contains the
stem plots and the second one the waterfall plots. They are identifyable
by unique id's which depend on which combination of merging the spectra into
one figure is used.


**Arguments:**

*--Required--*

- **files**:

    List of paths to the spectrum files. The files need to be given
    without their '.lin'/'.amps[xy]','.freqs[xy]' endings.  (So usually
    the path of the TbT-Data file.)


*--Optional--*

- **amp_limit** *(float)*:

    All amplitudes <= limit are filtered. This value needs to be at least
    0 to filter non-found frequencies.

    default: ``0.0``


- **bpms**:

    List of BPMs for which spectra will be plotted. If not given all BPMs
    are used.


- **combine_by**:

    Choose how to combine the data into figures.

    choices: ``['bpms', 'files']``

    default: ``[]``


- **filetype** *(str)*:

    Filetype to save plots as (i.e. extension without ".")

    default: ``pdf``


- **lines_manual** *(DictAsString)*:

    List of manual lines to plot. Need to contain arguments for axvline,
    and may contain the additional keys "text" and "loc" which is one of
    ['bottom', 'top', 'line bottom', 'line top'] and places the text at
    the given location.

    default: ``[]``


- **lines_nattunes** *(tuple)*:

    List of natural tune lines to plot

    default: ``[(1, 0), (0, 1)]``


- **lines_tunes** *(tuple)*:

    list of tune lines to plot

    default: ``[(1, 0), (0, 1)]``


- **manual_style** *(DictAsString)*:

    Additional style rcParameters which update the set of predefined ones.

    default: ``{}``


- **ncol_legend** *(int)*:

    Number of bpm legend-columns. If < 1 no legend is shown.

    default: ``5``


- **output_dir** *(str)*:

    Directory to write results to. If no option is given, plots will not
    be saved.


- **plot_styles** *(UnionPathStr)*:

    Which plotting styles to use, either from plotting.styles.*.mplstyles
    or default mpl.

    default: ``['standard', 'spectrum']``


- **plot_type**:

    Choose plot type (Multiple choices possible).

    choices: ``['stem', 'waterfall']``

    default: ``['stem']``


- **rescale**:

    Flag to rescale plots amplitude to max-line = 1

    action: ``store_true``


- **show_plots**:

    Flag to show plots

    action: ``store_true``


- **waterfall_cmap** *(str)*:

    Colormap to use for waterfall plot.

    default: ``inferno``


- **waterfall_common_plane_colors**:

    Same colorbar scale for both planes in waterfall plots.

    action: ``store_true``


- **waterfall_line_width**:

    Line width of the waterfall frequency lines. "auto" fills them up
    until the next one.

    default: ``2``


- **xlim** *(float)*:

    Limits on the x axis (Tupel)

    default: ``[0, 0.5]``


- **ylim** *(float)*:

    Limits on the y axis (Tupel)

    default: ``[1e-09, 1.0]``


"""
import os
from collections import OrderedDict
from typing import Tuple

from cycler import cycler
from generic_parser.entry_datatypes import DictAsString
from generic_parser.entrypoint_parser import (entrypoint, EntryPointParameters,
                                              save_options_to_config, DotDict)
from matplotlib import cm

from omc3.definitions import formats
from omc3.plotting.spectrum.stem import create_stem_plots
from omc3.plotting.spectrum.utils import (NCOL_LEGEND, LIN,
                                          LOG,
                                          FigureCollector, get_unique_filenames,
                                          filter_amps, get_bpms, get_stem_id,
                                          get_waterfall_id, get_data_for_bpm,
                                          load_spectrum_data)
from omc3.plotting.spectrum.waterfall import create_waterfall_plots
from omc3.plotting.utils import style as pstyle
from omc3.plotting.utils.lines import VERTICAL_LINES_TEXT_LOCATIONS
from omc3.utils import logging_tools
from omc3.utils.iotools import UnionPathStr

LOG = logging_tools.getLogger(__name__)


def get_reshuffled_tab20c():
    """
    Reshuffel tab20c so that the colors change between next lines.
    Needs to be up here as it is used in ``DEFAULTS`` which is loaded early.
    """
    tab20c = cm.get_cmap('tab20c').colors
    out = [None] * 20
    step, chunk = 4, 5
    for idx in range(step):
        start = idx * chunk
        out[start:start + chunk] = tab20c[idx::step]
    return cycler(color=out)


DEFAULTS = DotDict(
    waterfall_cmap='inferno',
    ylim=[1e-9, 1 ** .2],
    xlim=[0, .5],
    filetype='pdf',
    waterfall_line_width=2,
    manual_style={
        u'axes.prop_cycle': get_reshuffled_tab20c(),
    }
)


def get_params():
    params = EntryPointParameters()
    params.add_parameter(name="files",
                         required=True,
                         nargs='+',
                         help=("List of paths to the spectrum files. The files need to be given"
                               " without their '.lin'/'.amps[xy]','.freqs[xy]' endings. "
                               " (So usually the path of the TbT-Data file.)"))
    params.add_parameter(name="output_dir",
                         type=str,
                         help='Directory to write results to. If no option is given, plots will not be saved.')
    params.add_parameter(name="bpms",
                         nargs='+',
                         help='List of BPMs for which spectra will be plotted. If not given all BPMs are used.')
    params.add_parameter(name="amp_limit",
                         type=float,
                         default=0.,
                         help='All amplitudes <= limit are filtered. '
                              'This value needs to be at least 0 to filter non-found frequencies.')
    params.add_parameter(name="rescale",
                         action="store_true",
                         help='Flag to rescale plots amplitude to max-line = 1')
    params.add_parameter(name="plot_type",
                         nargs="+",
                         choices=['stem', 'waterfall'],
                         default=['stem'],
                         help='Choose plot type (Multiple choices possible).')
    params.add_parameter(name="combine_by",
                         nargs="*",
                         choices=['bpms', 'files'],
                         default=[],
                         help='Choose how to combine the data into figures.')
    params.add_parameter(name="waterfall_line_width",
                         default=DEFAULTS.waterfall_line_width,
                         help='Line width of the waterfall frequency lines. "auto" fills them up until the next one.')
    params.add_parameter(name="waterfall_cmap",
                         type=str,
                         default=DEFAULTS.waterfall_cmap,
                         help="Colormap to use for waterfall plot.")
    params.add_parameter(name="waterfall_common_plane_colors",
                         action="store_true",
                         help="Same colorbar scale for both planes in waterfall plots.")
    params.add_parameter(name="show_plots",
                         action="store_true",
                         help='Flag to show plots')
    params.add_parameter(name="lines_tunes",
                         nargs="*",
                         type=tuple,
                         default=[(1, 0), (0, 1)],
                         help='list of tune lines to plot')
    params.add_parameter(name="lines_nattunes",
                         nargs="*",
                         type=tuple,
                         default=[(1, 0), (0, 1)],
                         help='List of natural tune lines to plot')
    params.add_parameter(name="lines_manual",
                         nargs="*",
                         default=[],
                         type=DictAsString,
                         help='List of manual lines to plot. Need to contain arguments for axvline, and may contain '
                              'the additional keys "text" and "loc" which is one of '
                              f'{list(VERTICAL_LINES_TEXT_LOCATIONS.keys())} and places the text at the given location.'
                         )
    params.add_parameter(name="xlim",
                         nargs=2,
                         type=float,
                         default=DEFAULTS.xlim,
                         help='Limits on the x axis (Tupel)')
    params.add_parameter(name="ylim",
                         nargs=2,
                         type=float,
                         default=DEFAULTS.ylim,
                         help='Limits on the y axis (Tupel)')
    params.add_parameter(name="ncol_legend",
                         type=int,
                         default=NCOL_LEGEND,
                         help='Number of bpm legend-columns. If < 1 no legend is shown.')
    params.add_parameter(name="filetype",
                         type=str,
                         default=DEFAULTS.filetype,
                         help='Filetype to save plots as (i.e. extension without ".")')
    params.add_parameter(name="plot_styles",
                         type=UnionPathStr,
                         nargs="+",
                         default=['standard', 'spectrum'],
                         help='Which plotting styles to use, either from plotting.styles.*.mplstyles or default mpl.'
                         )
    params.add_parameter(name="manual_style",
                         type=DictAsString,
                         default={},
                         help='Additional style rcParameters which update the set of predefined ones.'
                         )
 
    return params


# Main -------------------------------------------------------------------------


@entrypoint(get_params(), strict=True)
def main(opt):
    LOG.info("Starting spectrum plots.")
    if opt.output_dir is not None:
        _save_options_to_config(opt)

    opt = _check_opt(opt)
    pstyle.set_style(opt.plot_styles, opt.manual_style)
    stem_opt, waterfall_opt, sorting_opt = _sort_opt(opt)
    stem, waterfall = _sort_input_data(sorting_opt)

    if stem_opt.plot:
        create_stem_plots(stem.figs, stem_opt)

    if waterfall_opt.plot:
        create_waterfall_plots(waterfall.figs, waterfall_opt)

    return stem.fig_dict, waterfall.fig_dict


# Input ------------------------------------------------------------------------


def _check_opt(opt):
    if (opt.waterfall_line_width is not None and opt.waterfall_line_width != DEFAULTS.waterfall_line_width
            and 'waterfall' not in opt.plot_type):
        LOG.warning("Setting 'waterfall_line_width' option has no effect, "
                    "when waterfall plots are deactivated!")

    if (opt.waterfall_cmap is not None and opt.waterfall_cmap != DEFAULTS.waterfall_cmap
            and 'waterfall' not in opt.plot_type):
        LOG.warning("Setting 'waterfall_cmap' option has no effect, "
                    "when waterfall plots are deactivated!")

    if opt.amp_limit < 0:
        raise ValueError("The amplitude limit needs to be at least '0' "
                         "to filter for non-found frequencies.")

    style_dict = DEFAULTS.manual_style.copy()
    if opt.manual_style is not None:
        style_dict.update(opt.manual_style)
    opt.manual_style = style_dict

    return opt


def _sort_opt(opt):
    # lines structure
    lines = opt.get_subdict(('lines_tunes', 'lines_nattunes', 'lines_manual'))
    lines = _rename_dict_keys(lines, to_remove="lines_")
    for key, val in lines.items():
        if val is None:
            lines[key] = []

    # limits structure
    limits = opt.get_subdict(("xlim", "ylim"))

    # stem-plot options
    stem = opt.get_subdict(('ncol_legend',))
    stem['plot'] = 'stem' in opt.plot_type

    # waterfall-plot options
    waterfall = opt.get_subdict(('waterfall_line_width', 'waterfall_cmap',
                                 'waterfall_common_plane_colors', 'ncol_legend'))
    waterfall = _rename_dict_keys(waterfall, to_remove="waterfall_")
    waterfall['plot'] = 'waterfall' in opt.plot_type

    # needed in both
    for d in (stem, waterfall):
        d['show'] = opt['show_plots']
        d['limits'] = limits
        d['lines'] = lines

    # sorting options
    sort = opt.get_subdict(('filetype', 'files',
                            'bpms', 'output_dir',
                            'amp_limit', 'rescale'))
    sort.plot_stem = stem.plot
    sort.plot_waterfall = waterfall.plot
    sort.combine_by = frozenset(opt.combine_by)

    return stem, waterfall, sort


# Output ---

def _save_options_to_config(opt):
    os.makedirs(opt.output_dir, exist_ok=True)
    save_options_to_config(os.path.join(opt.output_dir, formats.get_config_filename(__file__)),
                           OrderedDict(sorted(opt.items()))
                           )


# Load Data --------------------------------------------------------------------


def _sort_input_data(opt: DotDict) -> Tuple[FigureCollector, FigureCollector]:
    """Load and sort input data by file and bpm and assign correct figure-containers."""
    LOG.debug("Sorting input data.")

    stem_figs = FigureCollector()
    waterfall_figs = FigureCollector()

    # Data Sorting
    for file_path, filename in get_unique_filenames(opt.files):
        filename = "_".join(filename)
        LOG.info(f"Loading data for file '{filename}'.")

        data = load_spectrum_data(file_path, opt.bpms)
        data = filter_amps(data, opt.amp_limit)
        bpms = _get_all_bpms(get_bpms(data[LIN], opt.bpms, filename))

        for collector, get_id_fun, active in ((stem_figs, get_stem_id, opt.plot_stem),
                                              (waterfall_figs, get_waterfall_id, opt.plot_waterfall)):
            if not active:
                continue

            for bpm in bpms:
                the_id = get_id_fun(filename, bpm,
                                    opt.output_dir, opt.combine_by, opt.filetype)
                collector.add_data_for_id(the_id, get_data_for_bpm(data, bpm, opt.rescale))
    return stem_figs, waterfall_figs


# Helper -----------------------------------------------------------------------


def _get_all_bpms(bpms_dict):
    """Returns a union of all bpms for both planes."""
    return set.union(*[set(v) for v in bpms_dict.values()])


def _rename_dict_keys(d, to_remove):
    for key in list(d.keys()):  # using list to copy keys
        d[key.replace(to_remove, "")] = d.pop(key)
    return d


# Script Mode ------------------------------------------------------------------


if __name__ == "__main__":
    main()



"""
Plot Spectrum - Utilities
-------------------------

Common functions and sorting functions for the spectrum plotter.
"""
import os
from collections import OrderedDict
from contextlib import suppress
from pathlib import Path
from typing import Iterable, Sized, Union

import matplotlib
import numpy as np
import pandas as pd
import tfs
from generic_parser import DotDict
from matplotlib import transforms, axes, pyplot as plt
from matplotlib.patches import Rectangle

from omc3.definitions.constants import PLANES
from omc3.harpy.constants import FILE_AMPS_EXT, FILE_FREQS_EXT, FILE_LIN_EXT, COL_NAME
from omc3.plotting.utils.annotations import get_fontsize_as_float
from omc3.plotting.utils.lines import VERTICAL_LINES_ALPHA, plot_vertical_line
from omc3.utils import logging_tools

LOG = logging_tools.getLogger(__name__)

STEM_LINES_ALPHA = 0.5
PATCHES_ALPHA = 0.2
LABEL_Y_SPECTRUM = 'Amplitude in {plane:s} [a.u]'
LABEL_Y_WATERFALL = 'Plane {plane:s}'
LABEL_X = 'Frequency [tune units]'
NCOL_LEGEND = 5  # number of columns in the legend
WATERFALL_FILENAME = "waterfall_spectrum"
STEM_FILENAME = "stem_spectrum"
AMPS = FILE_AMPS_EXT.format(plane='')
FREQS = FILE_FREQS_EXT.format(plane='')
LIN = FILE_LIN_EXT.format(plane='')


# Collector Classes ------------------------------------------------------------


class FigureContainer(object):
    """ Container for attaching additional information to one figure. """
    def __init__(self, path: str) -> None:
        self.fig, self.axes = plt.subplots(nrows=len(PLANES), ncols=1)
        self.data = OrderedDict()  # make sure in plotting to use this order
        self.tunes = {p: [] for p in PLANES}
        self.nattunes = {p: [] for p in PLANES}
        self.path = path
        self.minmax = {p: (1, 0) for p in PLANES}

    def add_data(self, label: str, new_data: dict) -> None:
        self.data[label] = new_data
        for plane in PLANES:
            # Add tunes
            try:
                self.tunes[plane].append(new_data[plane][LIN].loc[f'TUNE{plane.upper()}'])
            except KeyError:
                LOG.warning(f'TUNE{plane.upper()} not found for {label}.')

            try:
                self.nattunes[plane].append(new_data[plane][LIN].loc[f'NATTUNE{plane.upper()}'])
            except KeyError:
                LOG.debug(f'NATTUNE{plane.upper()} not found for {label}.')

            # update min/max
            mmin, mmax = self.minmax[plane]
            self.minmax[plane] = (
                min(mmin, new_data[plane][AMPS].min(skipna=True)),
                max(mmax, new_data[plane][AMPS].max(skipna=True))
            )


class IdData:
    """ Container to keep track of the id-sorting output """
    def __init__(self, id_: str, label: str, path: str) -> None:
        self.id = id_        # id for the figure-container dictionary
        self.label = label  # plot labels
        self.path = path    # figure output path


class FigureCollector:
    """ Class to collect figure containers and manage data adding. """
    def __init__(self) -> None:
        self.fig_dict = {}   # dictionary of matplotlib figures, for output
        self.figs = {}       # dictionary of FigureContainers, used internally

    def add_data_for_id(self, id_data: IdData, data: dict) -> None:
        """ Add the data at the appropriate figure container. """
        try:
            figure_cont = self.figs[id_data.id]
        except KeyError:
            figure_cont = FigureContainer(id_data.path)
            self.figs[id_data.id] = figure_cont
            self.fig_dict[id_data.id] = figure_cont.fig
        figure_cont.add_data(id_data.label, data)


# (Tune-) Line Plotting --------------------------------------------------------


def plot_lines(fig_cont: FigureContainer, lines: DotDict) -> None:
    label_size = get_fontsize_as_float(matplotlib.rcParams['axes.labelsize']) * 0.7
    bottom_qlabel = 1.01

    for idx_plane, plane in enumerate(PLANES):
        ax = fig_cont.axes[idx_plane]
        trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
        bottom_natqlabel = bottom_qlabel + 2 * get_approx_size_in_axes_coordinates(ax, label_size)

        # Tune Lines ---
        for line_params in (("", fig_cont.tunes, lines.tunes, "--", bottom_qlabel),
                            ("NAT", fig_cont.nattunes, lines.nattunes, ":", bottom_natqlabel)):
            _plot_tune_lines(ax, trans, label_size, *line_params)

        # Manual Lines ---
        for mline in lines.manual:
            loc = mline.pop('loc', None)  # needs to be removed in axvline
            text = mline.pop('text', None)
            plot_vertical_line(ax, mline,  text, loc, label_size)
            mline['loc'] = loc  # reset it for later axes/plots
            mline['text'] = text


def _plot_tune_lines(ax, transform, label_size, q_string, tunes, resonances, linestyle, label_y):
    if len(resonances) == 0:
        return

    if all(len(tunes[p]) == 0 for p in PLANES):
        LOG.warning(f"Resonance lines can't be plotted for {q_string}, "
                    "as no tunes were found in files.")
        return

    pref = q_string[0] if len(q_string) else ""
    q_mean = _get_evaluated_tune_array(np.mean, tunes)
    q_min = _get_evaluated_tune_array(np.min, tunes)
    q_max = _get_evaluated_tune_array(np.max, tunes)
    freqs_mean = _get_resonance_frequencies(resonances, q_mean)
    freqs_min = _get_resonance_frequencies(resonances, q_min)
    freqs_max = _get_resonance_frequencies(resonances, q_max)
    for res, f_mean, f_min, f_max in zip(resonances, freqs_mean, freqs_min, freqs_max):
        if not np.isnan(f_mean):
            label, order = f'{pref}({res[0]}, {res[1]})', sum(np.abs(res)) + 1
            color = get_cycled_color(order-2)
            ax.axvline(x=f_mean, label=label,
                       linestyle=linestyle, color=color, marker='None',
                       zorder=-1, alpha=VERTICAL_LINES_ALPHA)
            ax.text(x=f_mean, y=label_y, s=label, transform=transform,
                    color=color,
                    va='bottom', ha='center',
                    fontdict={'size': label_size})
            ax.add_patch(Rectangle(xy=(f_min, 0), width=f_max-f_min, height=1,
                                   transform=transform, color=color, alpha=PATCHES_ALPHA, zorder=-2,))


def _get_resonance_frequencies(resonances, q):
    """ Calculates the frequencies for the resonance lines,
    but also filters lines in case the tune was not found. """
    resonances = np.array(resonances)

    # find zero-tune filter:
    # if tune in plane is not used (i.e. coefficient is zero) we can still plot the line
    use_idx = np.ones(resonances.shape[0], dtype=bool)
    for idx, tune in enumerate(q):
        if tune == 0:
            use_idx &= resonances[:, idx] == 0

    if sum(use_idx) == 0:
        LOG.warning("No usable tunes found to calculate resonance frequencies. "
                    " Maybe you gave natural lines for a free kick?")

    freqs = np.mod(resonances @ q, 1)
    freqs = np.where(freqs > .5, 1 - freqs, freqs)

    freqs.dtype = np.float64  # in case of all zeros, this is int and causes crash with float-nan
    freqs[~use_idx] = np.nan
    return freqs


def _get_evaluated_tune_array(fun, tunes):
    """ Array of tunes per plane that evaluates the tunes by fun,
    returns 0 where no tunes are present.
    """
    return np.array([fun(tunes[p]) if len(tunes[p]) else 0 for p in PLANES])


# ID Finder --------------------------------------------------------------------


def get_stem_id(filename: str, bpm: str, output_dir: str, combine_by: frozenset, filetype: str) -> IdData:
    """ Returns the stem-dictionary id and the path to which the output file should be written.
    By using more or less unique identifiers, this controls the creation of figures in the dictionary."""

    fun_map = {
        _fset("bpms", "files"): _get_id_single_fig_files_and_bpms,
        _fset("files"): _get_id_single_fig_files,
        _fset("bpms"): _get_id_single_fig_bpms,
        _fset(): _get_id_multi_fig,
    }
    return fun_map[combine_by](
        output_dir, STEM_FILENAME, filename, bpm, filetype
    )


def get_waterfall_id(filename: str, bpm: str, output_dir: str, combine_by: frozenset, filetype: str) -> IdData:
    """ Returns the waterfall-dictionary id and the path to which the output file should be written.
    By using identifiers for figures and unique lables per figure,
    this controls the creation of figures in the dictionary."""
    fun_map = {
        _fset("bpms", "files"): _get_id_single_fig_files_and_bpms,
        _fset("files"): _get_id_single_fig_files,
        _fset("bpms"): _get_id_single_fig_bpms,
        _fset(): _get_id_single_fig_bpms,  # same as above as single figure per file AND
    }                                      # bpm does not make sense for waterfall
    return fun_map[combine_by](
        output_dir, WATERFALL_FILENAME, filename, bpm, filetype
    )


def _fset(*args):
    """ Frozen Set shortcut for dict-key readability"""
    return frozenset(args)

# Specific Mappings ---


def _get_id_single_fig_files_and_bpms(output_dir: str, default_name: str, filename: str,
                                      bpm: str, filetype: str) -> IdData:
    """ Same id for all plots. Creates single figure.
    The label of the lines is a combination of filename and bpm.
    """
    return IdData(
        id_=default_name,
        label=f"{filename} {bpm}",
        path=_get_figure_path(output_dir, filename=None,
                              figurename=f"{default_name}.{filetype}")
    )


def _get_id_single_fig_files(output_dir: str, default_name: str, filename: str,
                             bpm: str, filetype: str) -> IdData:
    """ BPM as id for plots.
    Creates len(bpm) figures, with filenames as labels for lines.
    """
    return IdData(
        id_=bpm,
        label=filename,
        path=_get_figure_path(output_dir, filename=None,
                              figurename=f"{default_name}_{bpm}.{filetype}")
    )


def _get_id_single_fig_bpms(output_dir: str, default_name: str, filename: str,
                            bpm: str, filetype: str) -> IdData:
    """ Filename as ID for plots.
    Creates len(files) figures, with bpms as lables for lines.
    """
    return IdData(id_=filename,
                  label=bpm,
                  path=_get_figure_path(output_dir, filename=filename,
                                        figurename=f"{default_name}.{filetype}")
                  )


def _get_id_multi_fig(output_dir: str, default_name: str, filename: str,
                      bpm: str, filetype: str) -> IdData:
    """ Combination of Filename and BPM as ID. Creates len(files)*len(bpms) plots.
    BPM-name is printed as label.
    """
    return IdData(id_=f"{filename}_{bpm}",
                  label=bpm,
                  path=_get_figure_path(output_dir, filename=filename,
                                        figurename=f"{default_name}_{bpm}.{filetype}")
                  )


# Data Sorting -----------------------------------------------------------------


def get_data_for_bpm(data: dict, bpm: str, rescale: bool) -> dict:
    """ Loads data from files and returns a dictionary (over planes) of a
    dictionary over the files containing the bpm data as pandas series. """
    data_series = {p: {} for p in PLANES}
    for plane in PLANES:
        try:
            freqs = data[FREQS][plane].loc[:, bpm]
            amps = data[AMPS][plane].loc[:, bpm]
            lin = data[LIN][plane].loc[bpm, :]
        except KeyError:  # bpm not in this plane
            data_series[plane] = None
        else:
            idxs_data = _get_valid_indices(amps, freqs)
            data_series[plane][LIN] = lin
            data_series[plane][FREQS] = freqs.loc[idxs_data]
            data_series[plane][AMPS] = amps.loc[idxs_data]
            if rescale:
                data_series[plane][AMPS] = rescale_amp(data_series[plane][AMPS])

            if any(data_series[plane][AMPS].isna()):
                raise Exception("NAN FOUND")
    return data_series


def get_unique_filenames(files: Union[Iterable, Sized]):
    """ Way too complicated method to assure unique dictionary names,
        by going backwards through the file-path until the names differ.
    """
    paths = [None] * len(files)
    names = [None] * len(files)
    parts = -1
    for idx, fpath in enumerate(files):
        fpath = Path(fpath)
        fname = _get_partial_filepath(fpath, parts)
        while fname in names:
            parts -= 1
            for idx_old in range(idx):
                names[idx_old] = _get_partial_filepath(paths[idx_old], parts)
            fname = _get_partial_filepath(fpath, parts)
        names[idx] = fname
        paths[idx] = fpath
    return zip(paths, names)


def _get_partial_filepath(path: Path, nparts: int):
    """ Returns the path from nparts until the end"""
    return path.parts[nparts:]


def _get_valid_indices(amps, freqs):
    """ Intersection of filtered AMPS and FREQS indices. """
    return index_filter(amps).intersection(index_filter(freqs))


def index_filter(data: pd.Series):
    """ Only non-NaN and non-Zero data allowed.
    (Amps should not be zero due to _filter_amps() anyway.)"""
    return data[~(data.isna() | (data == 0))].index


def filter_amps(files: dict, limit: float):
    """ Filter amplitudes by limit. """
    for plane in PLANES:
        filter_idx = files[AMPS][plane] <= limit
        files[AMPS][plane][filter_idx] = np.NaN
        files[FREQS][plane][filter_idx] = np.NaN
    return files


def get_bpms(lin_files: dict, given_bpms: Iterable, filename: str, planes: Iterable = PLANES) -> dict:
    """ Return the bpm-names of the given bpms as found in the lin files.
     'file_path' is only used for the error messages."""
    found_bpms = {}
    empty_planes = 0
    for plane in planes:
        found_bpms[plane] = list(lin_files[plane].index)
        if given_bpms is not None:
            found_bpms[plane] = _get_only_given_bpms(found_bpms[plane], given_bpms, plane, filename)

        if len(found_bpms[plane]) == 0:
            LOG.warning(f"(id:{filename}) No BPMs found for plane {plane}!")
            empty_planes += 1

    if empty_planes == len(planes):
        raise IOError(f"(id:{filename}) No BPMs found in any plane!")
    return found_bpms


def _get_only_given_bpms(found_bpms, given_bpms, plane, file_path):
    found_bpms = [bpm for bpm in found_bpms if bpm in given_bpms]
    missing_bpms = [bpm for bpm in given_bpms if bpm not in found_bpms]
    if len(missing_bpms):
        LOG.warning(
            f"({file_path}) The following BPMs are not present or not present in plane {plane}:"
            f" {list2str(missing_bpms)}"
        )
    return found_bpms


def rescale_amp(amp_data: pd.Series) -> pd.Series:
    # return amp_data.divide(amp_data.max(axis=0), axis=1)  # dataframe
    return amp_data.divide(amp_data.max(skipna=True))  # series


# For Output ---


def output_plot(fig_cont: FigureContainer):
    fig = fig_cont.fig
    if fig_cont.path is not None:
        LOG.info(f"Saving Plot '{fig_cont.path}'")
        fig.savefig(fig_cont.path)


def _get_figure_path(out_dir, filename, figurename):
    path = _make_output_dir(out_dir, filename)
    if path is not None and figurename is not None:
        path = os.path.join(path, figurename)
    return path


def _make_output_dir(out_dir, filename):
    if out_dir is not None:
        if filename is not None:
            out_dir = os.path.join(out_dir, os.path.splitext(filename)[0])
        os.makedirs(out_dir, exist_ok=True)
    return out_dir


# Plotting Helper --------------------------------------------------------------


def get_cycled_color(idx: int):
    """ Get the color at (wrapped) idx in the color cycle. The CN-Method only works until 'C9'."""
    cycle = matplotlib.rcParams[u"axes.prop_cycle"].by_key()['color']
    return cycle[idx % len(cycle)]


def get_approx_size_in_axes_coordinates(ax: axes.Axes, label_size: float) -> float:
    transform = ax.transAxes.inverted().transform
    _, label_size_ax = transform((0, label_size)) - transform((0, 0))
    return label_size_ax


def list2str(list_: list):
    return str(list_)[1:-1]


# Spectrum File Loading --------------------------------------------------------


def load_spectrum_data(file_path: Path, bpms: Iterable, planes: Iterable = PLANES):
    """ Load Amps, Freqs and Lin Files into a dictionary, keys are the fileendings without plane,
     with subdicts of the planes. """
    LOG.info("Loading HARPY data.")
    with suppress(FileNotFoundError):
        return _get_harpy_data(file_path, planes)

    LOG.info("Some files not present. Loading SUSSIX data format")
    with suppress(FileNotFoundError):
        return _get_sussix_data(file_path, bpms, planes)

    raise FileNotFoundError(f"Neither harpy nor sussix files found in '{file_path.parent}' "
                            f"matching the name '{file_path.name}'.")


# Harpy Data ---


def _get_harpy_data(file_path, planes):
    return {
        AMPS: _get_planed_files(file_path, ext=FILE_AMPS_EXT, planes=planes),
        FREQS: _get_planed_files(file_path, ext=FILE_FREQS_EXT, planes=planes),
        LIN: _get_planed_files(file_path, ext=FILE_LIN_EXT, planes=planes, index=COL_NAME),
    }


def _get_planed_files(file_path, ext, planes, index=None):
    return {
        plane: tfs.read(
            str(file_path.with_name(file_path.name + ext.format(plane=plane.lower()))),
            index=index)
        for plane in planes
    }


# SUSSIX Data ---


def _get_sussix_data(file_path, bpms, planes):
    bpm_dir = file_path.parent / 'BPM'
    files = {LIN: {}, AMPS: {}, FREQS: {}}
    for plane in planes:
        files[LIN][plane] = tfs.read(
            str(file_path.with_name(f'{file_path.name}_lin{plane.lower()}')),
            index=COL_NAME)
        for id_ in (FREQS, AMPS):
            files[id_][plane] = tfs.TfsDataFrame(columns=bpms)
        for bpm in bpms:
            with suppress(FileNotFoundError):
                df = tfs.read(str(bpm_dir / f'{bpm}.{plane.lower()}'))
                files[FREQS][plane][bpm] = df["FREQ"]
                files[AMPS][plane][bpm] = df["AMP"]
        for id_ in (FREQS, AMPS):
            files[id_][plane] = files[id_][plane].fillna(0)
    return files



"""
BBQ Tools
---------

Tools to handle BBQ data.
"""
from typing import Tuple, Union
from datetime import datetime

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray

from omc3.utils import logging_tools
from omc3.utils.outliers import get_filter_mask
from dataclasses import  dataclass


LOG = logging_tools.get_logger(__name__)


@dataclass
class OutlierFilterOpt:
    """Options for moving average with outlier filter.

    Args:
        window: length of the averaging window.
        limit: cleaning limit, i.e. points with deviating less than this limit from the average will not be cleaned
     """
    window: int
    limit: float = 0.


@dataclass
class MinMaxFilterOpt:
    """Options for moving average with min/max filter.

    Args:
        window: length of the averaging window.
        min: minimum value (for filtering).
        max: maximum value (for filtering).
        fine_window: length of the averaging window for fine cleaning.
        fine_cut: allowed deviation for fine cleaning.
    """
    window: int
    min: float = None
    max: float = None
    fine_window: int = None
    fine_cut: float = None


FilterOpts = Union[OutlierFilterOpt, Tuple[MinMaxFilterOpt, MinMaxFilterOpt]]


def get_moving_average(data_series, filter_opt: MinMaxFilterOpt):
    """
    Get a moving average of the ``data_series`` over ``length`` entries. The data can be filtered
    beforehand. The values are shifted, so that the averaged value takes ceil((length-1)/2)
    values previous and floor((length-1)/2) following values into account.

    Args:
        data_series: `Series` of data.
        filter_opt: Options for the filtering, see `:class:`omc3.tune_analysis.bbq_tools.MinMaxFilterOpt`.

    Returns:
        A filtered and averaged `Series` and the mask used for filtering data.
    """
    LOG.debug(f"Cutting data and calculating moving average of length {filter_opt.window:d}.")

    if bool(filter_opt.fine_window) != bool(filter_opt.fine_cut):
        raise NotImplementedError("To activate fine cleaning, both "
                                  "'fine_window' and 'fine_cut' are needed.")

    if filter_opt.min is not None:
        min_mask = data_series <= filter_opt.min
    else:
        min_mask = np.zeros(data_series.size, dtype=bool)

    if filter_opt.max is not None:
        max_mask = data_series >= filter_opt.max
    else:
        max_mask = np.zeros(data_series.size, dtype=bool)

    cut_mask = min_mask | max_mask | data_series.isna()
    _is_almost_empty_mask(~cut_mask, filter_opt.window)
    data_mav, std_mav = _get_interpolated_moving_average(data_series, cut_mask, filter_opt.window)

    if filter_opt.fine_window is not None:
        min_mask = data_series <= (data_mav - filter_opt.fine_cut)
        max_mask = data_series >= (data_mav + filter_opt.fine_cut)
        cut_mask = min_mask | max_mask | data_series.isna()
        _is_almost_empty_mask(~cut_mask, filter_opt.fine_window)
        data_mav, std_mav = _get_interpolated_moving_average(data_series, cut_mask, filter_opt.fine_window)

    return data_mav, std_mav, ~cut_mask


def clean_outliers_moving_average(data_series: pd.Series, filter_opt: OutlierFilterOpt) -> Tuple[pd.Series, pd.Series, NDArray[bool]]:
    """
    Get a moving average of the ``data_series`` over ``length`` entries, by means of
    :func:`outlier filter <omc3.utils.outliers.get_filter_mask>`.
    The values are shifted, so that the averaged value takes ceil((length-1)/2) values previous
    and floor((length-1)/2) following values into account.

    Args:
        data_series: `Series` of data.
        filter_opt: Options for the filtering, see :class:`omc3.tune_analysis.bbq_tools.OutlierFilterOpt`.
    """
    LOG.debug(f"Filtering and calculating moving average of length {filter_opt.window:d}.")
    window, limit = filter_opt.window, filter_opt.limit
    init_mask = ~data_series.isna()
    mask = init_mask.copy()
    for i in range(len(data_series)-filter_opt.window):
        mask[i:i+window] &= get_filter_mask(data_series[i:i+window], limit=limit, mask=init_mask[i:i+window])

    _is_almost_empty_mask(mask, window)
    data_mav, std_mav = _get_interpolated_moving_average(data_series, ~mask, window)
    return data_mav, std_mav, mask


# Private methods ############################################################


def _get_interpolated_moving_average(data_series: pd.Series, clean_mask: Union[pd.Series, ArrayLike], length: int) -> Tuple[pd.Series, pd.Series]:
    """
    Returns the moving average of data series with a window of length and interpolated ``NaNs``.
    """
    data = data_series.copy()
    is_datetime_index = isinstance(data.index[0], datetime)

    if is_datetime_index:
        # in case data_series has datetime or similar as index
        # interpolation works in some pandas/numpy combinations, in some not
        try:
            # if clean_mask is a Series, bring into the right order and make array ...
            clean_mask = clean_mask[data.index].to_numpy()
        except TypeError:
            pass

        # ... as we change the index of data now
        data.index = pd.Index([i.timestamp() for i in data.index])

    data[clean_mask] = np.NaN

    try:
        # 'interpolate' fills nan based on index/values of neighbours
        data = data.interpolate("index").fillna(method="bfill").fillna(method="ffill")
    except TypeError as e:
        raise TypeError("Interpolation failed. "
                        "Usually due to a dtype format that is not properly recognized.") from e

    shift = -int((length-1)/2)  # Shift average to middle value

    # calculate mean and std, fill NaNs at the ends
    data_mav = data.rolling(length).mean().shift(shift).fillna(
        method="bfill").fillna(method="ffill")
    std_mav = data.rolling(length).std().shift(shift).fillna(
        method="bfill").fillna(method="ffill")

    if is_datetime_index:
        # restore index from input series
        data_mav.index = data_series.index
        std_mav.index = data_series.index

    return data_mav, std_mav


def _is_almost_empty_mask(mask, av_length):
    """Checks if masked data could be used to calculate moving average."""
    if sum(mask) <= av_length:
        raise ValueError("Too many points have been filtered. Maybe wrong filtering parameters?")



"""
Model Creator
-------------

Entrypoint to run the model creator for LHC, PSBooster and PS models.
"""
from pathlib import Path

from generic_parser import EntryPointParameters, entrypoint

from omc3.madx_wrapper import run_string
from omc3.model import manager
from omc3.model.accelerators.accelerator import Accelerator
from omc3.model.constants import JOB_MODEL_MADX
from omc3.model.model_creators.lhc_model_creator import (  # noqa
    LhcBestKnowledgeCreator,
    LhcCouplingCreator,
    LhcModelCreator,
)
from omc3.model.model_creators.ps_model_creator import PsModelCreator
from omc3.model.model_creators.psbooster_model_creator import PsboosterModelCreator
from omc3.model.model_creators.segment_creator import SegmentCreator
from omc3.utils.iotools import create_dirs
from omc3.utils import logging_tools

LOG = logging_tools.get_logger(__name__)


CREATORS = {
    "lhc": {"nominal": LhcModelCreator,
            "best_knowledge": LhcBestKnowledgeCreator,
            "segment": SegmentCreator,
            "coupling_correction": LhcCouplingCreator},
    "psbooster": {"nominal": PsboosterModelCreator,
                  "segment": SegmentCreator},
    "ps": {"nominal": PsModelCreator,
           "segment": SegmentCreator},
}


def _get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="type",
        choices=("nominal", "best_knowledge", "coupling_correction"),
        help="Type of model to create."
    )
    params.add_parameter(
        name="outputdir",
        required=True,
        type=Path,
        help="Output path for model, twiss files will be writen here."
    )
    params.add_parameter(
        name="logfile",
        type=Path,
        help=("Path to the file where to write the MAD-X script output."
              "If not provided it will be written to sys.stdout.")
    )
    return params


# Main functions ###############################################################


@entrypoint(_get_params())
def create_instance_and_model(opt, accel_opt) -> Accelerator:
    """
    Manager Keyword Args:
        *--Required--*

        - **accel**:

            Choose the accelerator to use.Can be the class already.

            choices: ``['lhc', 'ps', 'esrf', 'psbooster', 'skekb', 'JPARC', 'petra', 'iota']``


    Creator Keyword Args:
        *--Required--*

        - **outputdir** *(str)*:

            Output path for model, twiss files will be writen here.


        *--Optional--*

        - **logfile** *(str)*:

            Path to the file where to write the MAD-X script output.If not
            provided it will be written to sys.stdout.


        - **type**:

            Type of model to create.

            choices: ``('nominal', 'best_knowledge', 'coupling_correction')``


    Accelerator Keyword Args:
        lhc: :mod:`omc3.model.accelerators.lhc`

        ps: :mod:`omc3.model.accelerators.ps`

        esrf: :mod:`omc3.model.accelerators.esrf`

        psbooster: :mod:`omc3.model.accelerators.psbooster`

        skekb: :mod:`omc3.model.accelerators.skekb`

        iota: :mod:`omc3.model.accelerators.iota`

        petra: :mod:`omc3.model.accelerators.petra` (not implemented)

        JPARC: Not implemented
    """
    # Prepare paths
    create_dirs(opt.outputdir)

    accel_inst = manager.get_accelerator(accel_opt)
    LOG.info(f"Accelerator Instance {accel_inst.NAME}, model type {opt.type}")
    creator = CREATORS[accel_inst.NAME][opt.type]

    # Prepare model-dir output directory
    accel_inst.model_dir = opt.outputdir
    creator.prepare_run(accel_inst)

    # get madx-script with relative output-paths
    # as `cwd` changes run to correct directory.
    # The resulting model-dir is then more self-contained. (jdilly)
    accel_inst.model_dir = Path()
    madx_script = creator.get_madx_script(accel_inst)

    # Run madx to create model
    run_string(madx_script,
               output_file=opt.outputdir / JOB_MODEL_MADX,
               log_file=opt.logfile,
               cwd=opt.outputdir)

    # Check output and return accelerator instance
    accel_inst.model_dir = opt.outputdir
    creator.check_run_output(accel_inst)
    return accel_inst


if __name__ == "__main__":
    create_instance_and_model()



"""
Combined RDTs
-------------

This module contains combined resonance driving terms calculations functionality of
``optics_measurements``.
It provides functions to compute combined resonance driving terms following the derivations in
https://arxiv.org/pdf/1402.1461.pdf.
"""
from pathlib import Path
import numpy as np
import pandas as pd
import tfs
import scipy.odr
from omc3.optics_measurements.constants import ERR, EXT, AMPLITUDE, MDL, PHASE, REAL, IMAG
from omc3.utils import iotools, logging_tools
from omc3.utils.stats import circular_nanmean, circular_nanerror
from omc3.definitions.constants import PLANES, PI2
from omc3.harpy.constants import COL_AMP, COL_PHASE, COL_ERR
from omc3.optics_measurements.rdt import get_line_sign_and_suffix

LOGGER = logging_tools.get_logger(__name__)

CRDT_COLUMNS = [AMPLITUDE, f'{ERR}{AMPLITUDE}', PHASE, f'{ERR}{PHASE}']


CRDTS = [
    {'order': "skew_quadrupole", 'term': "F_XY", 'plane': 'X', 'line': [0, 1]},
    {'order': "skew_quadrupole", 'term': "F_YX", 'plane': 'Y', 'line': [1, 0]},

    {'order': "normal_sextupole", 'term': "F_NS3", 'plane': 'X', 'line': [-2, 0]},
    {'order': "normal_sextupole", 'term': "F_NS2", 'plane': 'X', 'line': [0, -2]},
    {'order': "normal_sextupole", 'term': "F_NS1", 'plane': 'Y', 'line': [-1, -1]},
    {'order': "normal_sextupole", 'term': "F_NS0", 'plane': 'Y', 'line': [1, -1]},

    {'order': "skew_sextupole", 'term': "F_SS3", 'plane': 'Y', 'line': [0, -2]},
    {'order': "skew_sextupole", 'term': "F_SS2", 'plane': 'Y', 'line': [-2, 0]},
    {'order': "skew_sextupole", 'term': "F_SS1", 'plane': 'X', 'line': [-1, -1]},
    {'order': "skew_sextupole", 'term': "F_SS0", 'plane': 'X', 'line': [1, -1]},

    {'order': "normal_octupole", 'term': "F_NO5", 'plane': 'Y', 'line': [0, 3]},
    {'order': "normal_octupole", 'term': "F_NO4", 'plane': 'X', 'line': [1, 2]},
    {'order': "normal_octupole", 'term': "F_NO3", 'plane': 'X', 'line': [3, 0]},
    {'order': "normal_octupole", 'term': "F_NO2", 'plane': 'X', 'line': [-1, 2]},
    {'order': "normal_octupole", 'term': "F_NO1", 'plane': 'Y', 'line': [2, -1]},
    {'order': "normal_octupole", 'term': "F_NO0", 'plane': 'Y', 'line': [2, 1]},
]


def calculate(measure_input, input_files, invariants, header):

    LOGGER.info("Start of CRDT analysis")
    assert len(input_files['X']) == len(input_files['Y'])
    bpm_names = input_files.bpms(dpp_value=0)
    for crdt in CRDTS:
        LOGGER.debug(f"Processing CRDT {crdt['term']}")
        result_df = generic_dataframe(input_files, measure_input, bpm_names)
        phase_sign, line_suffix = get_line_sign_and_suffix(crdt["line"], input_files, crdt["plane"])
        lines_and_phases = get_column_names(line_suffix)
        nqx, nqy = crdt['line']
        crdt_order = abs(nqx) + abs(nqy)
        # don't know the exact way to get to this via line indices so for now only via hacky way
        signflip = -1 if crdt['term'] in ("F_NO2", "F_NO1") else 1

        result_df = pd.merge(result_df, input_files.joined_frame(crdt["plane"], lines_and_phases.values()),
                             how='inner', left_index=True, right_index=True)

        amplitudes = input_files.get_data(result_df, lines_and_phases[AMPLITUDE])
        err_amplitudes = input_files.get_data(result_df, lines_and_phases[f"{ERR}{AMPLITUDE}"])
        result_df[AMPLITUDE], result_df[f'{ERR}{AMPLITUDE}'] = get_crdt_amplitude(crdt, invariants, amplitudes, err_amplitudes)

        phases = ((phase_sign * input_files.get_data(result_df, lines_and_phases[PHASE]) -
                  nqx * input_files.get_data(result_df, "MUX") -
                  nqy * input_files.get_data(result_df, "MUY") -
                  signflip * 0.75 - 0.5 * ((crdt_order)//3)) % 1)
        err_phases = np.sqrt(input_files.get_data(result_df, lines_and_phases[f'{ERR}{PHASE}'])**2 +
                             (nqx * input_files.get_data(result_df, "ERRMUX"))**2 +
                             (nqy * input_files.get_data(result_df, "ERRMUY"))**2)

        result_df[PHASE] = circular_nanmean(phases, axis=1, errors=err_phases)
        result_df[f'{ERR}{PHASE}'] = circular_nanerror(phases, axis=1, errors=err_phases)

        result_df[REAL] = np.cos(PI2 * result_df[PHASE].to_numpy()) * result_df[AMPLITUDE].to_numpy()
        result_df[IMAG] = np.sin(PI2 * result_df[PHASE].to_numpy()) * result_df[AMPLITUDE].to_numpy()

        write(result_df. loc[:, ["S", AMPLITUDE, f'{ERR}{AMPLITUDE}', PHASE, f'{ERR}{PHASE}', REAL, IMAG]],
              add_line_and_freq_to_header(header, crdt), measure_input, crdt['order'], crdt['term'])


def generic_dataframe(input_files, measure_input, bpm_names):
    result_df = pd.DataFrame(measure_input.accelerator.model).loc[bpm_names, ["S", "MUX", "MUY"]]
    result_df.rename(columns={f"MUX": f"MUX{MDL}", f"MUY": f"MUY{MDL}"}, inplace=True)
    for plane in PLANES:
        result_df = pd.merge(
            result_df, input_files.joined_frame(plane, [f"MU{plane}", f"{ERR}MU{plane}"], dpp_value=0),
            how='inner', left_index=True, right_index=True)
    return result_df


def add_line_and_freq_to_header(header, crdt):
    mod_header = header.copy()
    mod_header["LINE"] = f"{crdt['plane']}({crdt['line'][0]}, {crdt['line'][1]})"
    freq = np.mod(crdt['line']@np.array([header['Q1'], header['Q2']]), 1)
    mod_header["FREQ"] = freq if freq <= 0.5 else 1 - freq
    return mod_header


def write(df, header, meas_input, order, crdt):
    outputdir = Path(meas_input.outputdir)/"crdt"/order
    iotools.create_dirs(outputdir)
    tfs.write(str(outputdir/f"{crdt}{EXT}"), df, header, save_index='NAME')


def get_column_names(line):
    return dict(zip(CRDT_COLUMNS, [f'{COL_AMP}{line}',
                                   f'{COL_ERR}{COL_AMP}{line}',
                                   f'{COL_PHASE}{line}',
                                   f'{COL_ERR}{COL_PHASE}{line}']))


def get_crdt_invariant(crdt, invariants):
    exp = {'X': np.abs(crdt['line'][0]), 'Y': np.abs(crdt['line'][1])}
    exp[crdt['plane']] = exp[crdt['plane']] - 1 # to compensate for the normalization with tune line

    crdt_invariant = invariants['X'].T[0]**exp['X'] * invariants['Y'].T[0]**exp['Y']
    err_crdt_invariant = np.sqrt((exp['X'] * crdt_invariant / invariants['X'].T[0])**2 *
                                 invariants['X'].T[1]**2 +
                                 (exp['Y'] * crdt_invariant / invariants['Y'].T[0])**2 *
                                 invariants['Y'].T[1]**2)
    return crdt_invariant, err_crdt_invariant


def get_crdt_amplitude(crdt, invariants, line_amps, line_amp_errors):
    crdt_invariants, err_crdt_invariants = get_crdt_invariant(crdt, invariants)
    amps, err_amps = np.zeros(line_amps.shape[0]), np.zeros(line_amps.shape[0])

    for idx in range(line_amps.shape[0]):
        amps[idx], err_amps[idx] = fit_amplitude(line_amps[idx],
                                                 line_amp_errors[idx],
                                                 crdt_invariants,
                                                 err_crdt_invariants)

    return amps, err_amps


def fit_amplitude(lineamplitudes, err_lineamplitudes, crdt_invariant, err_crdt_invariant):

    def fun(p, x):
        return p * x * 2  # factor 2 to get to complex amplitudes again

    fit_model = scipy.odr.Model(fun)
    data_model = scipy.odr.RealData(
        x=crdt_invariant,
        y=lineamplitudes,
        sx=err_crdt_invariant,
        sy=None if all(err == 0. for err in err_lineamplitudes) else err_lineamplitudes,
    )
    odr = scipy.odr.ODR(data_model, fit_model, beta0=[0.5 * lineamplitudes[0] / crdt_invariant[0]])
    odr_output = odr.run()
    return odr_output.beta[0], odr_output.sd_beta[0]



"""
Response MAD-X
--------------

Provides a function to create the responses of beta, phase, dispersion, tune and coupling via iterative
madx calls.

The variables under investigation need to be provided as a list (which can be obtained from the accelerator
class).

For now, the response matrix is stored in a hdf5 file.

:author: Lukas Malina, Joschua Dilly, Jaime (...) Coello de Portugal
"""
import copy
import multiprocessing
from pathlib import Path
from typing import Dict, Sequence, Tuple, List

import numpy as np
import pandas as pd
import tfs
from optics_functions.coupling import coupling_via_cmatrix

import omc3.madx_wrapper as madx_wrapper
from omc3.correction.constants import (BETA, DISP, F1001, F1010, INCR,
                                       NORM_DISP, PHASE_ADV, TUNE, PHASE)
from omc3.model.accelerators.accelerator import Accelerator, AccElementTypes
from omc3.utils import logging_tools
from omc3.utils.contexts import suppress_warnings, timeit

LOG = logging_tools.get_logger(__name__)


# Full Response Mad-X ##########################################################


def create_fullresponse(
    accel_inst: Accelerator,
    variable_categories: Sequence[str],
    delta_k: float = 2e-5,
    num_proc: int = multiprocessing.cpu_count(),
    temp_dir: Path = None
) -> Dict[str, pd.DataFrame]:
    """ Generate a dictionary containing response matrices for
        beta, phase, dispersion, tune and coupling and saves it to a file.

        Args:
            accel_inst : Accelerator Instance.
            variable_categories (list): Categories of the variables/knobs to use. (from .json)
            delta_k (float): delta K1L to be applied to quads for sensitivity matrix
            num_proc (int): Number of processes to use in parallel.
            temp_dir (str): temporary directory. If ``None``, uses folder of original_jobfile.
    """
    LOG.debug("Generating Fullresponse via Mad-X.")
    with timeit(lambda t: LOG.debug(f"  Total time generating fullresponse: {t} s")):
        if not temp_dir:
            temp_dir = Path(accel_inst.model_dir)
        temp_dir.mkdir(parents=True, exist_ok=True)

        variables = accel_inst.get_variables(classes=variable_categories)
        if len(variables) == 0:
            raise ValueError("No variables found! Make sure your categories are valid!")
        num_proc = num_proc if len(variables) > num_proc else len(variables)
        process_pool = multiprocessing.Pool(processes=num_proc)

        incr_dict = _generate_madx_jobs(accel_inst, variables, delta_k, num_proc, temp_dir)
        _call_madx(process_pool, temp_dir, num_proc)
        _clean_up(temp_dir, num_proc)

        var_to_twiss = _load_madx_results(variables, process_pool, incr_dict, temp_dir)
        fullresponse = _create_fullresponse_from_dict(var_to_twiss)

    return fullresponse


def _generate_madx_jobs(
    accel_inst: Accelerator,
    variables: Sequence[str],
    delta_k: float,
    num_proc: int,
    temp_dir: Path
) -> Dict[str, float]:
    """ Generates madx job-files """
    LOG.debug("Generating MADX jobfiles.")
    incr_dict = {'0': 0.0}
    vars_per_proc = int(np.ceil(len(variables) / num_proc))

    madx_job = _get_madx_job(accel_inst)

    for proc_idx in range(num_proc):
        jobfile_path = _get_jobfiles(temp_dir, proc_idx)

        current_job = madx_job
        for i in range(vars_per_proc):
            var_idx = proc_idx * vars_per_proc + i
            if var_idx >= len(variables):
                break
            var = variables[var_idx]
            incr_dict[var] = delta_k
            current_job += f"{var}={var}{delta_k:+f};\n"
            current_job += f"twiss, file='{str(temp_dir / f'twiss.{var}')}';\n"
            current_job += f"{var}={var}{-delta_k:+f};\n\n"

        if proc_idx == num_proc - 1:
            current_job += f"twiss, file='{str(temp_dir / 'twiss.0')}';\n"

        jobfile_path.write_text(current_job)
    return incr_dict


def _get_madx_job(accel_inst: Accelerator) -> str:
    model_dir_backup = accel_inst.model_dir  # use relative paths as we use model_dir as cwd
    accel_inst.model_dir = Path()
    job_content = accel_inst.get_base_madx_script()
    accel_inst.model_dir = model_dir_backup
    job_content += (
        "select, flag=twiss, clear;\n"
        f"select, flag=twiss, pattern='{accel_inst.RE_DICT[AccElementTypes.BPMS]}', "
        "column=NAME,S,BETX,ALFX,BETY,ALFY,DX,DY,DPX,DPY,X,Y,K1L,MUX,MUY,R11,R12,R21,R22;\n\n")
    return job_content


def _call_madx(process_pool: multiprocessing.Pool, temp_dir: str, num_proc: int) -> None:
    """ Call madx in parallel """
    LOG.debug(f"Starting {num_proc:d} MAD-X jobs...")
    madx_jobs = [_get_jobfiles(temp_dir, index) for index in range(num_proc)]
    process_pool.map(_launch_single_job, madx_jobs)
    LOG.debug("MAD-X jobs done.")


def _clean_up(temp_dir: Path, num_proc: int) -> None:
    """ Merge Logfiles and clean temporary outputfiles """
    LOG.debug("Cleaning output and building log...")
    full_log = ""
    for index in range(num_proc):
        job_path = _get_jobfiles(temp_dir, index)
        log_path = job_path.with_name(f"{job_path.name}.log")
        full_log += log_path.read_text()
        log_path.unlink()
        job_path.unlink()
    full_log_path = temp_dir / "response_madx_full.log"
    full_log_path.write_text(full_log)


def _load_madx_results(
    variables: List[str],
    process_pool: multiprocessing.Pool,
    incr_dict: dict,
    temp_dir: Path
) -> Dict[str, tfs.TfsDataFrame]:
    """ Load the madx results in parallel and return var-tfs dictionary """
    LOG.debug("Loading Madx Results.")
    vars_and_paths = []
    for value in variables + ['0']:
        vars_and_paths.append((value, temp_dir))
    var_to_twiss = {}
    for var, tfs_data in process_pool.map(_load_and_remove_twiss, vars_and_paths):
        tfs_data[INCR] = incr_dict[var]
        var_to_twiss[var] = tfs_data
    return var_to_twiss


def _create_fullresponse_from_dict(var_to_twiss: Dict[str, tfs.TfsDataFrame]) -> Dict[str, pd.DataFrame]:
    """ Convert var-tfs dictionary to fullresponse dictionary. """
    var_to_twiss = _add_coupling(var_to_twiss)
    keys = list(var_to_twiss.keys())

    columns = [f"{PHASE_ADV}X", f"{PHASE_ADV}Y", f"{BETA}X", f"{BETA}Y", f"{DISP}X", f"{DISP}Y",
               f"{F1001}R", f"{F1001}I", f"{F1010}R", f"{F1010}I", f"{TUNE}1", f"{TUNE}2", INCR]

    bpms = var_to_twiss["0"].index
    resp = np.empty((len(keys), bpms.size, len(columns)))
    
    for i, key in enumerate(keys):
        resp[i] = var_to_twiss[key].loc[:, columns].to_numpy()

    resp = resp.transpose(2, 1, 0)
    model_index = list(keys).index("0")

    # create normalized dispersion and dividing BET by nominal model
    NDX_arr = np.divide(resp[columns.index(f"{DISP}X")], np.sqrt(resp[columns.index(f"{BETA}X")]))
    NDY_arr = np.divide(resp[columns.index(f"{DISP}Y")], np.sqrt(resp[columns.index(f"{BETA}Y")]))
    resp[columns.index(f"{BETA}X")] = np.divide(
        resp[columns.index(f"{BETA}X")], resp[columns.index(f"{BETA}X"), :, model_index][:, np.newaxis]
    )
    resp[columns.index(f"{BETA}Y")] = np.divide(
        resp[columns.index(f"{BETA}Y")], resp[columns.index(f"{BETA}Y"), :, model_index][:, np.newaxis]
    )

    # subtracting nominal model from data
    resp = np.subtract(resp, resp[:, :, model_index][:, :, np.newaxis])
    NDX_arr = np.subtract(NDX_arr, NDX_arr[:, model_index][:, np.newaxis])
    NDY_arr = np.subtract(NDY_arr, NDY_arr[:, model_index][:, np.newaxis])
    
    # Remove difference of nominal model with itself (bunch of zeros) and divide by increment
    resp = np.delete(resp, model_index, axis=2)
    NDX_arr = np.delete(NDX_arr, model_index, axis=1)
    NDY_arr = np.delete(NDY_arr, model_index, axis=1)
    keys.remove("0")

    NDX_arr = np.divide(NDX_arr, resp[columns.index(f"{INCR}")])
    NDY_arr = np.divide(NDY_arr, resp[columns.index(f"{INCR}")])
    resp = np.divide(resp,resp[columns.index(f"{INCR}")])
    Q_arr = np.column_stack((resp[columns.index(f"{TUNE}1"), 0, :], resp[columns.index(f"{TUNE}2"), 0, :])).T
 
    with suppress_warnings(np.ComplexWarning):  # raised as everything is complex-type now
        return {
            f"{PHASE_ADV}X": pd.DataFrame(data=resp[columns.index(f"{PHASE_ADV}X")], index=bpms, columns=keys).astype(np.float64),
            f"{PHASE_ADV}Y": pd.DataFrame(data=resp[columns.index(f"{PHASE_ADV}Y")], index=bpms, columns=keys).astype(np.float64),
            f"{BETA}X": pd.DataFrame(data=resp[columns.index(f"{BETA}X")], index=bpms, columns=keys).astype(np.float64),
            f"{BETA}Y": pd.DataFrame(data=resp[columns.index(f"{BETA}Y")], index=bpms, columns=keys).astype(np.float64),
            f"{DISP}X": pd.DataFrame(data=resp[columns.index(f"{DISP}X")], index=bpms, columns=keys).astype(np.float64),
            f"{DISP}Y": pd.DataFrame(data=resp[columns.index(f"{DISP}Y")], index=bpms, columns=keys).astype(np.float64),
            f"{NORM_DISP}X": pd.DataFrame(data=NDX_arr, index=bpms, columns=keys).astype(np.float64),
            f"{NORM_DISP}Y": pd.DataFrame(data=NDY_arr, index=bpms, columns=keys).astype(np.float64),
            f"{F1001}R": pd.DataFrame(data=resp[columns.index(f"{F1001}R")], index=bpms, columns=keys).astype(np.float64),
            f"{F1001}I": pd.DataFrame(data=resp[columns.index(f"{F1001}I")], index=bpms, columns=keys).astype(np.float64),
            f"{F1010}R": pd.DataFrame(data=resp[columns.index(f"{F1010}R")], index=bpms, columns=keys).astype(np.float64),
            f"{F1010}I": pd.DataFrame(data=resp[columns.index(f"{F1010}I")], index=bpms, columns=keys).astype(np.float64),
            f"{TUNE}": pd.DataFrame(data=Q_arr, index=[f"{TUNE}1", f"{TUNE}2"], columns=keys).astype(np.float64),
        }


def _get_jobfiles(temp_dir: Path, index: int) -> Path:
    """ Return names for jobfile and iterfile according to index """
    return temp_dir / f"job.iterate.{index:d}.madx"


def _launch_single_job(inputfile_path: Path) -> None:
    """ Function for pool to start a single madx job """
    log_file = inputfile_path.with_name(inputfile_path.name + ".log")
    madx_wrapper.run_file(inputfile_path, log_file=log_file, cwd=inputfile_path.parent)


def _load_and_remove_twiss(var_and_path: Tuple[str, Path]) -> Tuple[str, tfs.TfsDataFrame]:
    """ Function for pool to retrieve results """
    (var, path) = var_and_path
    twissfile = path / f"twiss.{var}"
    tfs_data = tfs.read(twissfile, index="NAME")
    tfs_data[f"{TUNE}1"] = tfs_data.Q1
    tfs_data[f"{TUNE}2"] = tfs_data.Q2
    twissfile.unlink()
    return var, tfs_data


def _add_coupling(dict_of_tfs: Dict[str, tfs.TfsDataFrame]) -> Dict[str, tfs.TfsDataFrame]:
    """
    For each TfsDataFrame in the input dictionary, computes the coupling RDTs and adds a column for
    the real and imaginary parts of the computed coupling RDTs. Returns a copy of the input dictionary with
    the aforementioned computed columns added for each TfsDataFrame.

    Args:
        dict_of_tfs (Dict[str, tfs.TfsDataFrame]): dictionary of Twiss dataframes.

    Returns:
        An identical dictionary of Twiss dataframes, with the computed columns added.
    """
    result_dict_of_tfs = copy.deepcopy(dict_of_tfs)
    with timeit(lambda elapsed: LOG.debug(f"  Time adding coupling: {elapsed} s")):
        for var, tfs_dframe in result_dict_of_tfs.items():  # already copies, so it's safe to act on them
            coupling_rdts_df = coupling_via_cmatrix(tfs_dframe)
            tfs_dframe[f"{F1001}R"] = np.real(coupling_rdts_df[f"{F1001}"]).astype(np.float64)
            tfs_dframe[f"{F1001}I"] = np.imag(coupling_rdts_df[f"{F1001}"]).astype(np.float64)
            tfs_dframe[f"{F1010}R"] = np.real(coupling_rdts_df[f"{F1010}"]).astype(np.float64)
            tfs_dframe[f"{F1010}I"] = np.imag(coupling_rdts_df[f"{F1010}"]).astype(np.float64)
        return result_dict_of_tfs



"""
Plot Amplitude Detuning Results
-------------------------------

Provides the plotting function for amplitude detuning analysis

**Arguments:**

*--Required--*

- **kicks**:

    Kick files as data frames or tfs files.


- **labels** *(str)*:

    Labels for the data. Needs to be same length as kicks.


- **plane** *(str)*:

    Plane of the kicks.

    choices: ``['X', 'Y', 'XY', '3D']``


*--Optional--*

- **action_plot_unit** *(str)*:

    Unit the action should be plotted in.

    choices: ``['km', 'm', 'mm', 'um', 'nm', 'pm', 'fm', 'am']``

    default: ``um``


- **action_unit** *(str)*:

    Unit the action is given in.

    choices: ``['km', 'm', 'mm', 'um', 'nm', 'pm', 'fm', 'am']``

    default: ``m``


- **correct_acd**:

    Correct for AC-Dipole kicks.

    action: ``store_true``


- **detuning_order** *(int)*:

    Order of the detuning as int. Basically just the order of the applied
    fit.

    default: ``1``


- **manual_style** *(DictAsString)*:

    Additional style rcParameters which update the set of predefined ones.

    default: ``{}``


- **output** *(str)*:

    Save the amplitude detuning plot here. Give filename with extension.
    An id for the 4 different plots will be added before the suffix.


- **plot_styles** *(UnionPathStr)*:

    Plotting styles.

    default: ``['standard', 'amplitude_detuning']``


- **show**:

    Show the amplitude detuning plot.

    action: ``store_true``


- **tune_scale** *(int)*:

    Plotting exponent of the tune.

    default: ``-3``


- **x_lim** *(float)*:

    Action limits in um (x-axis).


- **y_lim** *(float)*:

    Tune limits in units of tune scale (y-axis).


"""
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Dict, Sequence

import numpy as np
from generic_parser import entrypoint, EntryPointParameters, DotDict
from generic_parser.entry_datatypes import DictAsString
from matplotlib import colors as mcolors, MatplotlibDeprecationWarning
from matplotlib import pyplot as plt
from matplotlib import rcParams
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from numpy.typing import ArrayLike
from scipy import odr
from tfs.tools import significant_digits

from omc3.definitions.constants import UNIT_IN_METERS, PLANES
from omc3.plotting.utils import colors as pcolors, annotations as pannot, style as pstyle
from omc3.tune_analysis import constants as const, kick_file_modifiers as kick_mod, fitting_tools
from omc3.tune_analysis.kick_file_modifiers import AmpDetData
from omc3.utils import logging_tools
from omc3.utils.contexts import suppress_warnings
from omc3.utils.iotools import PathOrStr, save_config, UnionPathStr

LOG = logging_tools.get_logger(__name__)

NFIT = 100  # Points for the fitting function
X, Y = PLANES


def get_params():
    return EntryPointParameters(
        kicks=dict(
            nargs="+",
            help="Kick files as data frames or tfs files.",
            required=True,
        ),
        labels=dict(
            help="Labels for the data. Needs to be same length as kicks.",
            nargs='+',
            required=True,
            type=str,
        ),
        plane=dict(
            help="Plane of the kicks.",
            required=True,
            choices=list(PLANES) + [''.join(PLANES), '3D'],
            type=str,
        ),
        detuning_order=dict(
            help="Order of the detuning as int. Basically just the order of the applied fit.",
            type=int,
            default=1,
        ),
        correct_acd=dict(
            help="Correct for AC-Dipole kicks.",
            action="store_true",
        ),
        output=dict(
            help=("Save the amplitude detuning plot here. "
                  "Give filename with extension. An id for the 4 different "
                  "plots will be added before the suffix."),
            type=PathOrStr,
        ),
        show=dict(
            help="Show the amplitude detuning plot.",
            action="store_true",
        ),
        y_lim=dict(
            help="Tune limits in units of tune scale (y-axis).",
            type=float,
            nargs=2,
        ),
        x_lim=dict(
            help="Action limits in um (x-axis).",
            type=float,
            nargs=2,
        ),
        action_unit=dict(
            help="Unit the action is given in.",
            default="m",
            choices=list(UNIT_IN_METERS.keys()),
            type=str,
        ),
        action_plot_unit=dict(
            help="Unit the action should be plotted in.",
            default="um",
            choices=list(UNIT_IN_METERS.keys()),
            type=str,
        ),
        manual_style=dict(
            type=DictAsString,
            default={},
            help='Additional style rcParameters which update the set of predefined ones.'
        ),
        plot_styles=dict(
            help="Plotting styles.",
            type=UnionPathStr,
            nargs="+",
            default=['standard', 'amplitude_detuning'],
        ),
        tune_scale=dict(
            help="Plotting exponent of the tune.",
            default=-3,
            type=int,
        )
    )


@entrypoint(get_params(), strict=True)
def main(opt):
    LOG.info("Plotting Amplitude Detuning Results.")
    _save_options(opt)
    _check_opt(opt)

    figs = {}

    pstyle.set_style(opt.plot_styles, opt.manual_style)

    for tune_plane in PLANES:
        if opt.plane == "3D":
            figs.update(_plot_3d(tune_plane, opt))
        else:
            figs.update(_plot_2d(tune_plane, opt))

    if opt.show:
        plt.show()

    return figs


# Plotting --------------------------------------------------------------

# 2D Plots ----------------------------

def _plot_2d(tune_plane: str, opt: DotDict):
    """ 2D Plots per kick-plane and with/without BBQ correction. """
    figs = {}
    limits = opt.get_subdict(['x_lim', 'y_lim'])
    tune_scale = 10 ** opt.tune_scale

    for action_plane in opt.plane:
        do_acd_correction = opt.correct_acd and (action_plane == tune_plane)
        for corrected in [False, True]:  # with and without BBQ correction
            corr_label = "_corrected" if corrected else ""

            fig = plt.figure()
            ax = fig.add_subplot(111)
            isempty = True

            for idx, (kick_file, label) in enumerate(zip(opt.kicks, opt.labels)):
                kick_df = kick_mod.read_timed_dataframe(kick_file) if isinstance(kick_file, PathOrStr) else kick_file
                try:
                    data = kick_mod.get_ampdet_data(kick_df,
                                                    tune_plane=tune_plane,
                                                    action_plane=action_plane,
                                                    corrected=corrected
                                                    )
                except KeyError as e:
                    if corrected:
                        continue  # should only happen when there is no 'corrected' columns
                    raise

                # Read data from kick_df headers ---
                odr_fit = kick_mod.get_odr_data(kick_df,
                                                tune_plane=tune_plane,
                                                action_plane=action_plane,
                                                order=opt.detuning_order,
                                                corrected=corrected
                                                )

                # Get label for ODR fit ---
                odr_label = _get_odr_label(odr_fit,
                                           tune_plane=tune_plane,
                                           action_plane=action_plane,
                                           action_unit=opt.action_unit,
                                           do_acd_correction=do_acd_correction
                                           )

                # Scale and Plot ---
                data, odr_fit = _scale_data(
                    data=data, odr_fit=odr_fit,
                    action_unit=opt.action_unit, action_plot_unit=opt.action_plot_unit,
                    tune_scale=tune_scale
                )
                _plot_detuning(
                    ax, data=data, label=label, limits=limits,
                    odr_fit=odr_fit, odr_label=odr_label,
                    color=pcolors.get_mpl_color(idx),
                )
                isempty = False

            if isempty:
                plt.close(fig)  # don't show empty figs
                continue  # don't save/return empty figures

            if do_acd_correction:
                ax.text(0.0, 1.02, "* corrected for AC-Dipole.",
                        fontsize="x-small",
                        ha="left",
                        va='bottom',
                        transform=ax.transAxes,
                        )

            id_str = f"dQ{tune_plane.upper():s}d2J{action_plane.upper():s}{corr_label:s}"
            pannot.set_name(id_str, fig)

            ax_labels = (
                const.get_action_label(action_plane, opt.action_plot_unit),
                const.get_tune_label(tune_plane, opt.tune_scale)
            )
            _format_axes(ax, labels=ax_labels, limits=limits)

            if opt.output:
                output = Path(opt.output)
                fig.savefig(f'{output.with_suffix("")}_{id_str}{output.suffix}')

            figs[id_str] = fig
    return figs


def plot_odr(ax: Axes, odr_fit: odr.Output, xmax: float, label: str = '', color=None):
    """Adds a quadratic odr fit to axes."""

    color = 'k' if color is None else color
    odr_fit.beta[0] = 0   # no need to remove offset, as it is removed from data

    # get fits
    order = len(odr_fit.beta) - 1
    fit_fun = fitting_tools.get_poly_fun(order)
    f = partial(fit_fun, odr_fit.beta)
    f_low = partial(fit_fun, np.array(odr_fit.beta)-np.array(odr_fit.sd_beta))
    f_upp = partial(fit_fun, np.array(odr_fit.beta)+np.array(odr_fit.sd_beta))

    if order == 1:
        x = np.array([0, xmax])
    else:
        x = np.linspace(0, xmax, NFIT)

    line = ax.plot(x, f(x), marker="", linestyle='--', color=color, label=label)
    if color is None:
        color = line[0].get_color()
    ax.fill_between(x, f_low(x), f_upp(x), facecolor=mcolors.to_rgba(color, .3), zorder=-10)
    return color


def _plot_detuning(ax: Axes, data: AmpDetData, label: str, color=None,
                   limits: Dict[str, float] = None, odr_fit: odr.Output=None, odr_label: str =""):
    """Plot the detuning and the ODR into axes."""
    x_lim = _get_default(limits, 'x_lim', [0, max(data.action+data.action_err)])
    offset = 0

    # Plot Fit
    if odr_fit is not None:
        offset = odr_fit.beta[0]
        color = plot_odr(ax, odr_fit, xmax=x_lim[1], color=color, label=odr_label)

    # Plot Data
    data.tune -= offset
    ax.errorbar(x=data.action, xerr=data.action_err,
                y=data.tune, yerr=data.tune_err,
                label=label, color=color)


def _format_axes(ax: Axes, limits: Dict[str, Sequence[float]], labels: Sequence[str]):
    # labels
    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])

    # limits
    if limits['x_lim'] is not None:
        ax.set_xlim(limits['x_lim'])

    if limits['y_lim'] is not None:
        ax.set_ylim(limits['y_lim'])

    pannot.make_top_legend(ax, ncol=2)


# 3D Plots ----------------------------

def _plot_3d(tune_plane: str, opt: DotDict):
    """ 3D Plots for both kick-planes, one plot each with/without BBQ correction. """
    figs = {}
    limits = opt.get_subdict(['x_lim', 'y_lim'])
    tune_scale = 10 ** opt.tune_scale

    for corrected in [False, True]:  # with and without BBQ correction
        corr_label = "_corrected" if corrected else ""

        # hack to draw spines properly, as their style is
        # determined by lines
        rc_save = rcParams.copy()
        rcParams['lines.linestyle'] = "-"
        rcParams['lines.marker'] = ""

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        all_odr_labels = [None] * len(opt.kicks)

        # restore rcParams from before the hack
        with suppress_warnings(MatplotlibDeprecationWarning):
            rcParams.update(rc_save)
        isempty = True

        for idx, (kick_file, label) in enumerate(zip(opt.kicks, opt.labels)):
            kick_df = kick_mod.read_timed_dataframe(kick_file) if isinstance(kick_file, PathOrStr) else kick_file

            datas, odr_fits, odr_labels = {}, {}, {}
            for action_plane in PLANES:
                do_acd_correction = opt.correct_acd and (action_plane == tune_plane)
                try:
                    data = kick_mod.get_ampdet_data(kick_df,
                                                    action_plane=action_plane,
                                                    tune_plane=tune_plane,
                                                    corrected=corrected)
                except KeyError as e:
                    LOG.debug(str(e))
                    continue  # should only happen when there is no 'corrected' columns

                # Read data from kick_df headers ---
                odr_fit = kick_mod.get_odr_data(kick_df, action_plane, tune_plane,
                                                order=opt.detuning_order,
                                                corrected=corrected)

                # Get label for ODR fit ---
                odr_label = _get_odr_label(odr_fit,
                                           tune_plane=tune_plane,
                                           action_plane=action_plane,
                                           action_unit=opt.action_unit,
                                           do_acd_correction=do_acd_correction)

                # Scale ---
                data, odr_fit = _scale_data(
                    data, odr_fit,
                    opt.action_unit, opt.action_plot_unit,
                    tune_scale
                )

                # Save ---
                isempty = False
                odr_fits[action_plane] = odr_fit
                odr_labels[action_plane] = odr_label
                datas[action_plane] = data

            if len(datas) < 2:
                continue

            _plot_detuning_3d(
                ax, data=datas, label=label, limits=limits,
                odr_fits=odr_fits,
                color=pcolors.get_mpl_color(idx),
            )
            all_odr_labels[idx] = odr_labels

        if isempty:
            plt.close(fig)  # don't show empty figs
            continue  # don't save/return empty figures

        id_str = f"dQ{tune_plane.upper():s}d2JXY_3D{corr_label:s}"
        pannot.set_name(id_str, fig)

        ax_labels = (
            const.get_action_label(PLANES[0], opt.action_plot_unit),
            const.get_action_label(PLANES[1], opt.action_plot_unit),
            const.get_tune_label(tune_plane, opt.tune_scale)
        )
        ax.view_init(azim=45, elev=18)
        _format_axes_3d(ax, ax_labels=ax_labels, limits=limits,
                        acd=opt.correct_acd, odr_labels=all_odr_labels)

        if opt.output:
            output = Path(opt.output)
            fig.savefig(f'{output.with_suffix("")}_{id_str}{output.suffix}')
        figs[id_str] = fig
    return figs


def _plot_detuning_3d(ax: Axes, data: Dict[str, AmpDetData], label: str, color=None, limits=None, odr_fits=None):
    """ Plot the detuning and the ODR into axes."""
    offset = 0
    jx, jx_err = data[X].action, data[X].action_err
    jy, jy_err = data[Y].action, data[Y].action_err
    tune, tune_err = data[X].tune, data[X].tune_err
    x_lim = _get_default(limits, 'x_lim', [0, max(jx+jx_err)])
    y_lim = _get_default(limits, 'x_lim', [0, max(jy+jy_err)])

    # Plot Fit
    if odr_fits is not None:
        offset = odr_fits[X].beta[0]
        color = plot_odr_3d(ax, odr_fits, xymax=[x_lim[1], y_lim[1]], color=color)

    # Plot Data
    tune -= offset

    # there is no errorbar3D in mpl it seems
    # so plotting is done manually.
    # plot points
    line = ax.plot(jx, jy, tune, label=label, color=color)
    if color is None:
        color = line[0].get_color()

    # plot errorbars
    for idx in np.arange(0, len(tune)):
        x, dx = jx[idx], jx_err[idx]
        y, dy = jy[idx], jy_err[idx]
        z, dz = tune[idx], tune_err[idx]

        ax.plot([x+dx, x-dx], [y, y], [z, z], ls="-", marker="_", color=color)
        ax.plot([x, x], [y+dy, y-dy], [z, z], ls="-", marker="_", color=color)
        ax.plot([x, x], [y, y], [z+dz, z-dz], ls="-", marker="_", color=color)


@dataclass
class FitFuncs:
    fit: callable
    upper: callable
    lower: callable


def fit_fun_odr_1d(x, y, q0, qdx, qdy):
    return q0 + qdx * x + qdy * y


def plot_odr_3d(ax: Axes, odr_fits: Dict[str, odr.Output], xymax: Sequence[float], color=None):
    """Plot the odr fit in 3D."""

    color = 'k' if color is None else color
    odr_fits[X].beta[0] = 0   # no need to remove offset, as it is removed from data
    odr_fits[Y].beta[0] = 0   # no need to remove offset, as it is removed from data

    # get fits
    order = len(odr_fits[X].beta) - 1
    if order > 1:
        raise NotImplementedError("ODR fit plots for order > 1 are not implemented.")
    f = partial(fit_fun_odr_1d,
                q0=odr_fits[X].beta[0],
                qdx=odr_fits[X].beta[1],
                qdy=odr_fits[Y].beta[1])
    f_low = partial(fit_fun_odr_1d,
                    q0=odr_fits[X].beta[0]-odr_fits[X].sd_beta[0],
                    qdx=odr_fits[X].beta[1]-odr_fits[X].sd_beta[1],
                    qdy=odr_fits[Y].beta[1]-odr_fits[Y].sd_beta[1])
    f_upp = partial(fit_fun_odr_1d,
                    q0=odr_fits[X].beta[0]+odr_fits[X].sd_beta[0],
                    qdx=odr_fits[X].beta[1]+odr_fits[X].sd_beta[1],
                    qdy=odr_fits[Y].beta[1]+odr_fits[Y].sd_beta[1])

    x = np.linspace(0, xymax[0], 6)
    y = np.linspace(0, xymax[1], 6)
    xm, ym = np.meshgrid(x, y)

    # line = ax.plot_surface(xm, ym, f(xm, ym), color=color, label=label, alpha=0.3)
    frame = ax.plot_wireframe(xm, ym, f(xm, ym), color=color, alpha=0.5)
    if color is None:
        color = frame[0].get_color()

    # Plot only the lower and upper boundaries
    # ax.plot_surface(xm, ym, f_low(xm, ym), color=color, alpha=0.3)
    # ax.plot_surface(xm, ym, f_upp(xm, ym), color=color, alpha=0.3)

    # Plot the full surface
    x = np.linspace(0, xymax[0], 2)
    y = np.linspace(0, xymax[1], 2)
    plot_cube(ax, x, y, f_low, f_upp, color=color, alpha=0.2, rcount=2, ccount=2)
    return color


def plot_cube(ax: Axes, x: ArrayLike, y: ArrayLike, f_low: callable, f_upp: callable, **kwargs):
    """Plots a cube-like plot with f_low and f_upp as surfaces. """
    # lower
    xm, ym = np.meshgrid(x, y)
    zm = f_low(xm, ym)
    ax.plot_surface(xm, ym, zm, **kwargs)

    # upper
    zm = f_upp(xm, ym)
    ax.plot_surface(xm, ym, zm, **kwargs)

    # north
    xm, ym = np.array([x, x]), y[0]*np.ones([2, len(y)])
    zm = np.array([f_low(x, y[0]), f_upp(x, y[0])])
    ax.plot_surface(xm, ym, zm, **kwargs)

    # south
    xm, ym = np.array([x, x]), y[-1]*np.ones([2, len(y)])
    zm = np.array([f_low(x, y[-1]), f_upp(x, y[-1])])
    ax.plot_surface(xm, ym, zm, **kwargs)

    # east
    xm, ym = x[0]*np.ones([2, len(x)]), np.array([y, y])
    zm = np.array([f_low(x[0], y), f_upp(x[0], y)])
    ax.plot_surface(xm, ym, zm, **kwargs)

    # west
    xm, ym = x[-1]*np.ones([2, len(x)]), np.array([y, y])
    zm = np.array([f_low(x[-1], y), f_upp(x[-1], y)])
    ax.plot_surface(xm, ym, zm, **kwargs)


def _format_axes_3d(
        ax: Axes, limits: Dict[str, float], ax_labels: Sequence[str],
        acd: bool, odr_labels: Sequence[Dict[str, str]] = None):
    # labels
    ax.set_xlabel(ax_labels[0], labelpad=15)
    ax.set_ylabel(ax_labels[1], labelpad=15)
    ax.set_zlabel(ax_labels[2], labelpad=10)

    # limits
    if limits['x_lim'] is not None:
        limj = limits['x_lim']
    else:
        limj = [0, max([ax.get_xlim()[1], ax.get_ylim()[1]])]
    ax.set_xlim(limj)
    ax.set_ylim(limj)

    if limits['y_lim'] is not None:
        ax.set_zlim(limits['y_lim'])

    # pannot.make_top_legend(ax, ncol=2)
    if acd:
        ax.text2D(
            1.4, 0.02, "* corrected for AC-Dipole.",
            fontsize="x-small",
            ha="left",
            va='bottom',
            transform=ax.transAxes,
        )
    h, l = get_labels_with_odr_labels(ax, odr_labels)
    ax.legend(h, l, loc='upper left', bbox_to_anchor=(1.2, 0.98), prop={'size': 'small'})
    ax.zaxis._axinfo['juggled'] = (1, 2, 0)  # move tune axis to the other side
    # tight layout so that the legend fits
    ax.figure.tight_layout()
    ax.figure.tight_layout()


# Labels -----------------------------------------------------------------------

def _get_odr_label(odr_fit: odr.Output, tune_plane: str, action_plane: str,
                   action_unit: str, do_acd_correction: bool):
    """ Returns the label for the ODR fit, nicely formatted and scaled. """
    order = len(odr_fit.beta) - 1
    str_list = [None] * order
    for o in range(order):
         label = _get_scaled_odr_label(
            odr_fit,
            order=o+1,
            action_unit=action_unit,
            do_acd_correction=do_acd_correction,
            magnitude_exponent=const.get_detuning_exponent_for_order(o+1)
         )
         order_str = f"^{o+1}" if o else ""
         str_list[o] = f"$Q{order_str}_{{{tune_plane},{action_plane}}}$: {label}"
    return ", ".join(str_list)


def _get_scaled_odr_label(odr_fit, order, action_unit, do_acd_correction, magnitude_exponent=3):
    """ Returns the label for the ODR fit after scaling to readable units and accounting for AC-Dipole correction."""
    acd_scale = 0.5 if do_acd_correction else 1
    str_acd_scale = "$^*$" if do_acd_correction else ""

    scale = acd_scale * (10 ** -magnitude_exponent) / (UNIT_IN_METERS[action_unit] ** order)
    str_val, str_std = _get_scaled_labels(odr_fit.beta[order], odr_fit.sd_beta[order], scale)
    str_mag = ''
    if magnitude_exponent != 0:
        str_mag = fr'$\cdot$ 10$^{{{magnitude_exponent}}}$'
    return fr'({str_val} $\pm$ {str_std}){str_acd_scale} {str_mag} m$^{{-{order}}}$'


def _get_scaled_labels(val, std, scale):
    scaled_vas, scaled_std = val*scale, std*scale
    if abs(scaled_std) > 1 or scaled_std == 0:
        return f'{int(round(scaled_vas)):d}', f'{int(round(scaled_std)):d}'
    return significant_digits(scaled_vas, scaled_std)


def get_labels_with_odr_labels(ax, odr_labels):
    h, l = ax.get_legend_handles_labels()
    if odr_labels is None:
        return h, l

    empty = Line2D([0], [0], ls='none', marker='', label='')
    h_new, l_new = [], []
    for handle, label, odr in zip(h, l, odr_labels):
        h_new.append(handle)
        l_new.append(label)
        if odr is not None:
            h_new.append(empty)
            l_new.append("\n".join(odr.values()))
    return h_new, l_new


# Helper -----------------------------------------------------------------------

def _save_options(opt):
    if opt.output:
        out_path = Path(opt.output).parent
        save_config(out_path, opt, __file__)


def _check_opt(opt):
    if len(opt.labels) != len(opt.kicks):
        raise ValueError("'kicks' and 'labels' need to be of same size!")


def _get_default(ddict, key, default):
    """Returns ``default`` if either the dict itself or the entry is ``None``."""
    if ddict is None or key not in ddict or ddict[key] is None:
        return default
    return ddict[key]


def _scale_data(data: AmpDetData, odr_fit: odr.Output,
                action_unit: str, action_plot_unit: str, tune_scale: float):
    """Scale data to plot-units (y=tune_scale, x=um)."""
    x_scale = UNIT_IN_METERS[action_unit] / UNIT_IN_METERS[action_plot_unit]
    data.action *= x_scale
    data.action_err *= x_scale

    # correct for tune scaling
    y_scale = 1. / tune_scale
    data.tune *= y_scale
    data.tune_err *= y_scale

    # same for odr_fit:
    for idx in range(len(odr_fit.beta)):
        full_scale = y_scale / (x_scale**idx)
        odr_fit.beta[idx] *= full_scale
        odr_fit.sd_beta[idx] *= full_scale
    return data, odr_fit


# Script Mode ------------------------------------------------------------------


if __name__ == '__main__':
    main()




"""
Merge KMOD Results
------------------

Script to merge the results from KMOD into one `TfsDataFrame`.
The script takes the kmod-results folders as input and merges the lsa-result tfs-files in these
together.
The resulting `TfsDataFrame` is returned, but also written out if an ``outputdir`` is given.

**BPMWKs** (common for both beams) are hereby dropped, to avoid duplicated elements.

Headers of the same name will be overwritten (depending on the alphabetical order of the
directory names).

Some sanity checks are performed, e.g. that there is only one entry per element. If IP1 and IP5
results are given for both planes and beams, the luminosity imbalance between these IPs is also
calculated and written into the header, as well as logged.

**Arguments:**

*--Required--*

- **kmod_dirs** *(Path)*:

    Path to kmod directories with stored KMOD measurement files,in particular lsa_results.tfs


*--Optional--*

- **outputdir** *(Path)*:

    Output directory where to write the result tfs
"""
import pathlib
import re
from typing import Dict, List, Tuple

import numpy as np
import tfs
from generic_parser import EntryPointParameters, entrypoint
from uncertainties import UFloat, unumpy as up

from omc3.definitions.constants import PLANES
from omc3.kmod.constants import ERR, BETA, EXT
from omc3.kmod.constants import LSA_FILE_NAME as LSA_RESULTS
from omc3.utils.iotools import save_config
from omc3.utils.logging_tools import get_logger

LOG = get_logger(__name__)

NAME = "NAME"  # Column containing the IP/Beam names
BEAMS = ("B1", "B2")
IPS = ("ip1", "ip5")

HEADER_IMBALANCE = "LUMINOSITY_IMBALANCE"
HEADER_REL_ERROR = "RELATIVE_ERROR"
HEADER_EFF_BETA_IP1 = "EFFECTIVE_BETA_IP1"
HEADER_REL_ERROR_IP1 = "RELATIVE_ERROR_IP1"
HEADER_EFF_BETA_IP5 = "EFFECTIVE_BETA_IP5"
HEADER_REL_ERROR_IP5 = "RELATIVE_ERROR_IP5"

BPMWK = "BPMWK."  # elements starting with this are dropped


def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="kmod_dirs",
        type=pathlib.Path,
        nargs="+",
        required=True,
        help=f"Path to kmod directories with stored KMOD measurement files,"
             f"in particular {LSA_RESULTS}{EXT}",
    )
    params.add_parameter(
        name="outputdir",
        type=pathlib.Path,
        help="Output directory where to write the result tfs",
    )
    return params


@entrypoint(get_params(), strict=True)
def merge_kmod_results(opt) -> tfs.TfsDataFrame:
    """
    Main function to merge K-Mod output.
    See :mod:`omc3.scripts.merge_kmod_results`.
    """
    if opt.outputdir:
        save_config(opt.outputdir, opt, script=__file__)  # creates outputdir

    # Get the directories we need where the tfs are stored
    ip_dir_names: List[pathlib.Path] = get_kmod_ip_subdirs(opt.kmod_dirs)

    # Combine the lsa data
    lsa_tfs = merge_tfs(ip_dir_names, f"{LSA_RESULTS}{EXT}")

    # If the TFS data contains everything we need: get the imbalance
    if _validate_for_imbalance(lsa_tfs):
        imbalance_results = get_lumi_imbalance(lsa_tfs)
        lsa_tfs = _add_imbalance_to_header(lsa_tfs, *imbalance_results)

    if opt.outputdir:
        output_path = opt.outputdir / f"{LSA_RESULTS}{EXT}"
        LOG.info(f"Writing result TFS file to disk at: {str(output_path)}")
        tfs.write(output_path, lsa_tfs, save_index=NAME)
    return lsa_tfs


def _validate_for_imbalance(data_frame: tfs.TfsDataFrame) -> bool:
    """
    Checks that the provided `TfsDataFrame` contains the expected labels and columns for a
    luminosity imbalance calculation.

    Args:
        data_frame (tfs.TfsDataFrame): a loaded `TfsDataFrame` to validate.

    Returns:
        ``True`` if the provided dataframe is valid, ``False`` otherwise.
    """
    # check all required entries are there
    imbalance_names = [f"{ip}{beam}" for ip in IPS for beam in BEAMS]
    imbalance_columns = [f"{err}{BETA}{p}" for p in PLANES for err in ('', ERR)]
    try:
        data_frame.loc[imbalance_names, imbalance_columns]
    except KeyError:
        return False
    return True


def get_lumi_imbalance(data_frame: tfs.TfsDataFrame) -> Tuple[UFloat, UFloat, UFloat]:
    """
    Calculate the IP1 / IP5 luminosity imbalance. The luminosity is taken as defined in
    `Concept of luminosity`, Eq(17): https://cds.cern.ch/record/941318/files/p361.pdf

    The calculation of the luminosity imbalance is then:

    .. math::

        \\frac{L_{IP1}}{L_{IP5}}=\\frac{\\sqrt{\\beta_{x1,IP5}+\\beta_{x2,IP5}}
        \\cdot\\sqrt{\\beta_{y1,IP5}+\\beta_{y2,IP5}}}
        {\\sqrt{\\beta_{x1,IP1}+\\beta_{x2,IP1}}\\cdot\\sqrt{\\beta_{y1,IP1}+\\beta_{y2,IP1}}}

    and the error:

    .. math::

        \\sigma_{\\frac{L_{IP1}}{L_{IP5}}} = \\frac{1}{2}\\frac{L_{IP1}}{L_{IP5}}
        \\cdot \\sqrt{\\sum_{\substack{z \\in (x,y) \\\\ i \\in (IP1, IP5) }}
        {\\frac{\\sigma^2_{\\beta_{z1,i}} + \\sigma^2_{\\beta_{z2,i}}} {(\\beta_{z1,i}+\\beta_{z2,i})^2}}}

    Args:
        data_frame (tfs.TfsDataFrame): a `TfsDataFrame` with the results from a kmod analysis.

    Returns:
         Tuple with the imbalance, the betas at IPs as ufloats.
    """

    lumi_coefficient: Dict[str, UFloat] = {}

    for ip in IPS:
        LOG.debug(f"Computing luminosity contribution from optics for IP {ip}")
        ip_rows = data_frame.loc[[f"{ip}{beam}" for beam in BEAMS], :]

        beta_sums = [up.uarray(ip_rows[f"{BETA}{plane}"].to_numpy(),
                               ip_rows[f"{ERR}{BETA}{plane}"].to_numpy()
                               ).sum()  # sum over beams
                     for plane in PLANES]
        lumi_coefficient[ip] = 0.5 * up.sqrt(np.prod(beta_sums))

    imbalance = lumi_coefficient[IPS[0]] / lumi_coefficient[IPS[1]]

    LOG.info(f'{"Luminosity Imbalance":22s}: {imbalance:s}')
    LOG.info(f'{"Effective beta IP1":22s}: {lumi_coefficient[IPS[0]]:s}')
    LOG.info(f'{"Effective beta IP5":22s}: {lumi_coefficient[IPS[1]]:s}')

    return imbalance, lumi_coefficient[IPS[0]], lumi_coefficient[IPS[1]]


def _add_imbalance_to_header(
        tfs_df: tfs.TfsDataFrame,
        imbalance: UFloat,
        beta_ip1: UFloat,
        beta_ip5: UFloat
) -> tfs.TfsDataFrame:
    """
    Function to add the calculated imablance and effective betas to the header

    Args:
        tfs_df (tfs.TfsDataFrame): a `TfsDataFrame` with the results from a kmod analysis.
        imbalance (UFloat): uncertain imbalance.
        beta_ip1 (UFloat): uncertain effective beta in ip1.
        beta_ip5 (UFloat): uncertain effective beta in ip5.

    Returns:
        tfs.TfsDataFrame with added headers.
    """
    header_map = [(HEADER_IMBALANCE, HEADER_REL_ERROR, imbalance),
                  (HEADER_EFF_BETA_IP1, HEADER_REL_ERROR_IP1, beta_ip1),
                  (HEADER_EFF_BETA_IP5, HEADER_REL_ERROR_IP5, beta_ip5),
    ]
    for val_key, err_key, val in header_map:
        tfs_df.headers[val_key], tfs_df.headers[err_key] = str(val).split('+/-')

    return tfs_df


def merge_tfs(directories: List[pathlib.Path], filename: str) -> tfs.TfsDataFrame:
    """
    Merge different kmod analysis results from a list of directories into a single `TfsDataFrame`.

    Args:
        directories (List[pathlib.Path]): list of PosixPath objects to directories holding TFS
            files with the results of kmod analysis.
        filename (str): name of the TFS files to look for in the provided directories.

    Returns:
        A `TfsDataFrame` combining all the loaded files from the provided directories.
    """
    # Combine the data into one tfs
    new_tfs = tfs.TfsDataFrame()
    headers = {}
    for d in sorted(directories):
        loaded_tfs = tfs.read_tfs(d / filename)
        headers.update(loaded_tfs.headers)  # old headers are lost in `append`
        new_tfs = new_tfs.append(loaded_tfs, ignore_index=True)
    new_tfs.headers = headers
    new_tfs = new_tfs.set_index(NAME)

    # drop BPMWK and check tfs
    new_tfs = new_tfs.loc[~new_tfs.index.str.startswith(BPMWK), :]
    if not new_tfs.index.is_unique:
        raise KeyError("Found duplicated entries "
                       f"{', '.join(set(new_tfs.index[new_tfs.index.duplicated()]))}'.")
    return new_tfs


def get_kmod_ip_subdirs(kmod_dirs: List[pathlib.Path]) -> List[pathlib.Path]:
    """ Returns the paths to the ip-subdirectories in the kmod result directories.

    Args:
        kmod_dirs (List[Path]): Main kmod result directories with IP-subdirectories.

    Returns:
        (List[Path]) Paths to all subdirectories.
    """
    # Check directories first
    for main_dir in kmod_dirs:
        if not main_dir.is_dir():
            raise NotADirectoryError(f"Directory {main_dir} does not exist")

    pattern = re.compile(".*ip[0-9]B[1-2]")
    return [
        sub_dir for main_dir in kmod_dirs for sub_dir in main_dir.glob("**/*")
        if pattern.match(sub_dir.name) and sub_dir.is_dir()
    ]


if __name__ == "__main__":
    merge_kmod_results()



"""
Fake Measurement from Model
---------------------------

Script to generate a pseudo-measurement from a twiss-model.

The `model` the then generated measurement is compared to can be different
from the `twiss` given, e.g. if the `twiss` incorporates errors.


**Arguments:**

*--Required--*

- **twiss** *(PathOrStrOrDataFrame)*:

    Twiss dataframe or path to twiss-file.


*--Optional--*

- **model** *(PathOrStrOrDataFrame)*:

    Alternative Model (Dataframe or Path) to use. If not given, `twiss`
    will be used.


- **outputdir** *(PathOrStr)*:

    Path to the output directory for the fake measurement.


- **parameters** *(str)*:

    Optics parameters to use

    choices: ``('PHASEX', 'PHASEY', 'BETX', 'BETY', 'DX', 'DY', 'NDX', 'F1010', 'F1001')``

    default: ``['PHASEX', 'PHASEY', 'BETX', 'BETY', 'DX', 'DY', 'NDX', 'F1010', 'F1001']``


- **randomize** *(str)*:

    Randomize values and/or errors from gaussian distributions. If not
    randomized, measurement values will be equal to the model values and
    the errors will be equal to the relative error * measurement.

    choices: ``['values', 'errors']``

    default: ``['values', 'errors']``


- **relative_errors** *(float)*:

    Relative errors. Either single value for all paramters orlist of
    values in order of parameters.

    default: ``[0.0]``


- **seed** *(int)*:

    Set random seed.


"""
from pathlib import Path
from typing import Tuple, Sequence, List, Dict

import numpy as np
import pandas as pd
import tfs
from generic_parser import EntryPointParameters, entrypoint

from omc3.correction.constants import DISP, NORM_DISP, F1001, F1010
from omc3.correction.model_appenders import add_coupling_to_model
from omc3.definitions.constants import PLANES
from omc3.optics_measurements.constants import (BETA_NAME, AMP_BETA_NAME, PHASE_NAME,
                                                TOTAL_PHASE_NAME,
                                                DISPERSION_NAME, NORM_DISP_NAME,
                                                EXT, DELTA, ERR,
                                                PHASE_ADV, BETA, PHASE,
                                                TUNE, NAME, NAME2, S, MDL, REAL, IMAG)
from omc3.optics_measurements.toolbox import df_rel_diff, df_ratio, df_diff, df_ang_diff, ang_interval_check, ang_diff
from omc3.utils import logging_tools
from omc3.utils.iotools import PathOrStrOrDataFrame, PathOrStr

LOG = logging_tools.get_logger(__name__)

OUTPUTNAMES_MAP = {
    # Names to be output on input of certain parameters.
    f'{BETA}X': tuple(f"{name}x" for name in (BETA_NAME, AMP_BETA_NAME)),
    f'{BETA}Y': tuple(f"{name}y" for name in (BETA_NAME, AMP_BETA_NAME)),
    f'{DISP}X': tuple([f"{DISPERSION_NAME}x"]),
    f'{DISP}Y': tuple([f"{DISPERSION_NAME}y"]),
    f'{PHASE}X': tuple(f"{name}x" for name in (PHASE_NAME, TOTAL_PHASE_NAME)),
    f'{PHASE}Y': tuple(f"{name}y" for name in (PHASE_NAME, TOTAL_PHASE_NAME)),
    F1010: tuple([F1010.lower()]),
    F1001: tuple([F1001.lower()]),
    f'{NORM_DISP}X': tuple(f"{name}x" for name in (BETA_NAME, AMP_BETA_NAME, DISPERSION_NAME, NORM_DISP_NAME)),
}
FAKED_HEADER = "FAKED_FROM"
VALUES = 'values'
ERRORS = 'errors'
EPSILON = 1e-14  # smallest allowed relative error, empirical value
                 # could be smaller but then normal(mean, err) == mean sometimes


def get_params():
    params = EntryPointParameters()
    params.add_parameter(
        name="twiss",
        required=True,
        help="Twiss dataframe or path to twiss-file.",
        type=PathOrStrOrDataFrame,
    )
    params.add_parameter(
        name="model",
        help=("Alternative Model (Dataframe or Path) to use. "
             "If not given, `twiss` will be used."),
        type=PathOrStrOrDataFrame,
    )
    params.add_parameter(
        name="parameters",
        help="Optics parameters to use",
        choices=list(OUTPUTNAMES_MAP.keys()),
        default=list(OUTPUTNAMES_MAP.keys()),
        type=str,
        nargs="+",
    )
    params.add_parameter(
        name="relative_errors",
        help=("Relative errors. Either single value for all paramters or"
              "list of values in order of parameters."),
        default=[0.,],
        type=float,
        nargs="+",
    )
    params.add_parameter(
        name="randomize",
        help=("Randomize values and/or errors from gaussian distributions."
              " If not randomized, measurement values will be equal to the model "
              "values and the errors will be equal to the relative error * measurement."
              ),
        choices=[VALUES, ERRORS],
        default=[VALUES, ERRORS],
        type=str,
        nargs="*",
    )
    params.add_parameter(
        name="outputdir",
        help="Path to the output directory for the fake measurement.",
        type=PathOrStr,
    )
    params.add_parameter(
        name="seed",
        help="Set random seed.",
        type=int,
    )
    return params


@entrypoint(get_params(), strict=True)
def generate(opt) -> Dict[str, tfs.TfsDataFrame]:
    """
    Takes a twiss file and writes the parameters in optics_parameters to Output_dir in the format
    global_correction_entrypoint uses (same format you would get from hole_in_one).
    """
    LOG.info("Generating fake measurements.")
    # prepare data
    np.random.seed(opt.seed)
    randomize = opt.randomize if opt.randomize is not None else []
    df_twiss, df_model = _get_data(opt.twiss, opt.model,
                                   add_coupling=(F1001 in opt.parameters) or (F1010 in opt.parameters))

    # headers
    headers = {f"{TUNE}1": df_twiss[f"{TUNE}1"], f"{TUNE}2": df_twiss[f"{TUNE}2"]}
    if isinstance(opt.twiss, PathOrStr):
        headers[FAKED_HEADER] = str(opt.twiss)

    # create defaults
    results = {}
    for parameter, error in _get_loop_parameters(opt.parameters, opt.relative_errors):
        create = CREATOR_MAP[parameter[:-1]]
        new_dfs = create(df_twiss, df_model, parameter,
                         relative_error=error,
                         randomize=randomize,
                         headers=headers)
        results.update(new_dfs)

    # maybe create normalized dispersion
    for plane in PLANES:
        nd_param = f'{NORM_DISP}{plane}'
        if nd_param in opt.parameters:
            nd_df = create_normalized_dispersion(results[f'{DISPERSION_NAME}{plane.lower()}'],
                                                 results[f'{BETA_NAME}{plane.lower()}'],
                                                 df_model,
                                                 nd_param, headers)
            results.update(nd_df)

    # output
    if opt.outputdir is not None:
        LOG.info("Writing fake measurements to files.")
        output_path = Path(opt.outputdir)
        for filename, df in results.items():
            full_path = output_path / f"{filename}{EXT}"
            tfs.write(full_path, df, save_index=NAME)

    return results


# Main Creator Functions -------------------------------------------------------

def create_beta(df_twiss: pd.DataFrame, df_model: pd.DataFrame, parameter: str,
                relative_error: float, randomize: Sequence[str], headers: Dict):
    """ Create both beta_amp and beta_phase measurements. """
    LOG.info(f"Creating fake beta for {parameter}.")
    plane = parameter[-1]
    df = create_measurement(df_twiss, parameter, relative_error, randomize)
    df[parameter] = np.abs(df[parameter])

    df = append_model(df, df_model, parameter, plane, beat=True)
    df.headers = headers.copy()
    return {f'{BETA_NAME}{plane.lower()}': df,
            f'{AMP_BETA_NAME}{plane.lower()}': df}


def create_dispersion(df_twiss: pd.DataFrame, df_model: pd.DataFrame, parameter: str,
                      relative_error: float, randomize: Sequence[str], headers: Dict):
    """ Creates dispersion measurement. """
    LOG.info(f"Creating fake dispersion for {parameter}.")
    plane = parameter[-1]
    df = create_measurement(df_twiss, parameter, relative_error, randomize)
    df = append_model(df, df_model, parameter, plane)
    df.headers = headers.copy()
    return {f'{DISPERSION_NAME}{plane.lower()}': df}


def create_phase(df_twiss: pd.DataFrame, df_model: pd.DataFrame, parameter: str,
                 relative_error: float, randomize: Sequence[str], headers: Dict):
    """ Creates both phase advance and total phase measurements. """
    LOG.info(f"Creating fake phases for {parameter}.")
    results = {}
    dict_adv = create_phase_advance(df_twiss, df_model, parameter, relative_error, randomize, headers)
    results.update(dict_adv)

    dict_tot = create_total_phase(df_twiss, df_model, parameter, relative_error, randomize, headers)
    results.update(dict_tot)
    return results


def create_phase_advance(df_twiss: pd.DataFrame, df_model: pd.DataFrame, parameter: str,
                         relative_error: float, randomize: Sequence[str], headers: Dict):
    """ Creates phase advance measurements. """
    LOG.debug(f"Creating fake phase advance for {parameter}.")
    plane = parameter[-1]
    df_adv = tfs.TfsDataFrame(index=df_twiss.index[:-1])
    df_adv[NAME2] = df_twiss.index[1:].to_numpy()

    def get_phase_advances(df_source):
        return ang_diff(
                df_source.loc[df_adv[NAME2], f"{PHASE_ADV}{plane}"].to_numpy(),
                df_source.loc[df_adv.index, f"{PHASE_ADV}{plane}"].to_numpy()
        )

    values = get_phase_advances(df_twiss)
    errors = relative_error * np.ones_like(values)
    if ERRORS in randomize:
        errors = _get_random_errors(errors, np.ones_like(values)) % 0.5

    if VALUES in randomize:
        values = np.random.normal(values, errors)
        values = ang_interval_check(values)

    df_adv[parameter] = values
    df_adv[f'{ERR}{parameter}'] = errors

    # adv model
    df_adv[S] = df_model.loc[df_adv.index, S]
    df_adv[f'{S}2'] = df_model.loc[df_adv[NAME2], S].to_numpy()
    df_adv[f'{parameter}{MDL}'] = get_phase_advances(df_model)
    df_adv[f'{PHASE_ADV}{plane}{MDL}'] = df_model.loc[df_adv.index, f'{PHASE_ADV}{plane}']

    df_adv[f"{ERR}{DELTA}{parameter}"] = df_adv[f'{ERR}{parameter}']
    df_adv[f"{DELTA}{parameter}"] = df_ang_diff(df_adv, parameter, f'{parameter}{MDL}')
    df_adv.headers = headers.copy()
    return {f'{PHASE_NAME}{plane.lower()}': df_adv}


def create_total_phase(df_twiss: pd.DataFrame, df_model: pd.DataFrame, parameter: str,
                       relative_error: float, randomize: Sequence[str], headers: Dict):
    """ Creates total phase measurements. """
    LOG.debug(f"Creating fake total phase for {parameter}.")
    plane = parameter[-1]
    df_tot = tfs.TfsDataFrame(index=df_twiss.index)
    element0 = df_twiss.index[0]
    df_tot[NAME2] = element0

    values = df_twiss[f"{PHASE_ADV}{plane}"] - df_twiss.loc[element0, f"{PHASE_ADV}{plane}"]
    errors = relative_error * np.ones_like(values)
    if ERRORS in randomize:
        errors = _get_random_errors(errors, np.ones_like(values)) % 0.5
    errors[0] = 0.

    if VALUES in randomize:
        rand_val = np.random.normal(values, errors) % 1
        values += ang_diff(rand_val, values)

    df_tot[parameter] = values % 1
    df_tot[f'{ERR}{parameter}'] = errors

    # tot model
    df_tot[S] = df_model[S]
    df_tot[f'{S}2'] = df_model.loc[df_tot.index[0], S]
    df_tot[f'{parameter}{MDL}'] = (
                                          df_model[f"{PHASE_ADV}{plane}"]
                                          - df_model.loc[element0, f"{PHASE_ADV}{plane}"]
                                  ) % 1
    df_tot[f'{PHASE_ADV}{plane}{MDL}'] = df_model[f'{PHASE_ADV}{plane}']

    df_tot[f"{ERR}{DELTA}{parameter}"] = df_tot[f'{ERR}{parameter}']
    df_tot[f"{DELTA}{parameter}"] = df_ang_diff(df_tot, parameter, f'{parameter}{MDL}')
    df_tot = df_tot.fillna(0)
    df_tot.headers = headers.copy()
    return {f'{TOTAL_PHASE_NAME}{plane.lower()}': df_tot}


def create_coupling(df_twiss: pd.DataFrame, df_model: pd.DataFrame, parameter: str,
                    relative_error: float, randomize: Sequence[str], headers: Dict):
    """ Creates coupling measurements for either the difference or sum RDT. """
    LOG.info(f"Creating fake coupling for {parameter}.")
    # Naming with R and I as long as model is involved
    re = create_measurement(df_twiss, f'{parameter}R', relative_error, randomize)
    im = create_measurement(df_twiss, f'{parameter}I', relative_error, randomize)
    df = pd.concat([re, im], axis=1)
    df = append_model(df, df_model, f'{parameter}R')
    df = append_model(df, df_model, f'{parameter}I', planes='XY')

    # Go to RDT naming scheme
    df.columns = df.columns.str.replace(f'{parameter}R', REAL).str.replace(f'{parameter}I', IMAG)
    df.headers = headers.copy()
    return {parameter.lower(): df}


CREATOR_MAP = {
    BETA: create_beta,
    DISP: create_dispersion,
    PHASE: create_phase,
    F1010[:-1]: create_coupling,  # normally the plane is removed but here is no plane
    F1001[:-1]: create_coupling,
}


# Not mapped ---

def create_normalized_dispersion(df_disp: pd.DataFrame, df_beta: pd.DataFrame,
                                 df_model: pd.DataFrame, parameter: str, headers: Dict):
    """ Creates normalized dispersion from pre-created dispersion and beta dataframes. """
    LOG.info(f"Creating fake normalized dispersion for {parameter}.")
    plane = parameter[-1]  # 'X'

    # Measurement
    df = tfs.TfsDataFrame(index=df_disp.index)
    disp = df_disp.loc[:, f"{DISP}{plane}"]
    disp_err = df_disp.loc[:, f"{ERR}{DISP}{plane}"]
    beta = df_beta.loc[:, f"{BETA}{plane}"]
    beta_err = df_beta.loc[:, f"{ERR}{BETA}{plane}"]

    inv_beta = 1/beta
    df[parameter] = disp / np.sqrt(beta)
    df[f"{ERR}{parameter}"] = np.sqrt(
        0.25 * disp**2 * inv_beta**3 * beta_err**2 + inv_beta * disp_err**2
    )

    # Model
    df_model[f'{parameter}'] = df_disp[f"{DISP}{plane}{MDL}"] / np.sqrt(df_beta[f"{BETA}{plane}{MDL}"])
    df = append_model(df, df_model, parameter, plane)

    df.headers = headers.copy()
    output_name = f'{NORM_DISP_NAME}{plane.lower()}'
    return {output_name: df}


def create_measurement(df_twiss: pd.DataFrame, parameter: str, relative_error: float,
                       randomize: Sequence[str]) -> tfs.TfsDataFrame:
    """ Create a new measurement Dataframe from df_twiss from parameter. """
    LOG.debug(f"Creating fake measurement for {parameter}.")
    values = df_twiss.loc[:, parameter]
    errors = relative_error * values.abs()
    if all(values == 0):
        LOG.warning(f"All value for {parameter} are zero. "
                    f"Fake measurement will be zero as well.")
    else:
        if ERRORS in randomize:
            errors = _get_random_errors(errors, values)

        if VALUES in randomize:
            values = np.random.normal(values, errors)

    df = tfs.TfsDataFrame({parameter: values,
                           f"{ERR}{parameter}": errors},
                          index=df_twiss.index)
    return df


def append_model(df: pd.DataFrame, df_model: pd.DataFrame, parameter: str,
                 planes: str = '', beat: bool = False) -> pd.DataFrame:
    """ Add the values to the measurement. """
    LOG.debug(f"Appending model to fake measurement for {parameter}.")
    df[S] = df_model[S]

    for plane in planes:
        df[f'{PHASE_ADV}{plane}{MDL}'] = df_model[f'{PHASE_ADV}{plane}']

    df[f"{parameter}{MDL}"] = df_model[f'{parameter}']
    if beat:
        df[f"{DELTA}{parameter}"] = df_rel_diff(df, parameter, f"{parameter}{MDL}")
        df[f"{ERR}{DELTA}{parameter}"] = df_ratio(df, f"{ERR}{parameter}", f"{parameter}{MDL}")
    else:
        df[f"{DELTA}{parameter}"] = df_diff(df, f'{parameter}', f'{parameter}{MDL}')
        df[f"{ERR}{DELTA}{parameter}"] = df[f'{ERR}{parameter}']
    return df


# Other Functions --------------------------------------------------------------

def _get_data(twiss: tfs.TfsDataFrame, model: tfs.TfsDataFrame = None,
              add_coupling: bool = False) -> Tuple[tfs.TfsDataFrame, tfs.TfsDataFrame]:
    """ Get's the input data as TfsDataFrames. """
    LOG.debug("Loading data.")
    def try_reading(df_or_path):
        try:
            return tfs.read(df_or_path, index=NAME)
        except TypeError:
            return df_or_path

    twiss = try_reading(twiss)
    if add_coupling:
        twiss = add_coupling_to_model(twiss)

    if model is None:
        model = twiss.copy()
    else:
        model = try_reading(model)
        if add_coupling:
            model = add_coupling_to_model(model)
    return twiss, model


def _get_loop_parameters(parameters: Sequence[str], errors: Sequence[float]) -> List[str]:
    """ Special care for normalized dispersion"""
    parameters = list(parameters)
    if errors is None:
        errors = [0.]
    errors = list(errors)
    if len(errors) == 1:
        errors = errors * len(parameters)

    for plane in PLANES:
        nd_param = f'{NORM_DISP}{plane}'
        if nd_param in parameters:
            idx = parameters.index(nd_param)
            nd_error = errors[idx]
            del errors[idx]
            parameters.remove(nd_param)

            add_params = [p for p in (f'{DISP}{plane}', f'{BETA}{plane}') if p not in parameters]
            parameters += add_params
            errors += [nd_error/2] * len(add_params)  # not really knowable here
    return zip(parameters, errors)


def _get_random_errors(errors: np.array, values: np.array) -> np.array:
    """ Creates normal distributed error-values that will not be lower than EPSILON. """
    random_errors = np.zeros_like(errors)
    too_small = np.ones_like(errors, dtype=bool)
    while sum(too_small):
        random_errors[too_small] = np.random.normal(errors[too_small], errors[too_small])
        too_small = random_errors < EPSILON * np.abs(values)
    return random_errors


# Script Mode ------------------------------------------------------------------

if __name__ == '__main__':
    generate()
